#########################################################################
##   This file is part of the α,β-CROWN (alpha-beta-CROWN) verifier    ##
##                                                                     ##
##   Copyright (C) 2021-2024 The α,β-CROWN Team                        ##
##   Primary contacts: Huan Zhang <huan@huan-zhang.com>                ##
##                     Zhouxing Shi <zshi@cs.ucla.edu>                 ##
##                     Kaidi Xu <kx46@drexel.edu>                      ##
##                                                                     ##
##    See CONTRIBUTORS for all author contacts and affiliations.       ##
##                                                                     ##
##     This program is licensed under the BSD 3-Clause License,        ##
##        contained in the LICENCE file in this directory.             ##
##                                                                     ##
#########################################################################
"""Helper for the BaBSR-like heuristic on general nonlinear functions."""
import numpy as np

import torch

from collections import deque
from auto_LiRPA.backward_bound import get_degrees, add_constant_node, add_bound
from auto_LiRPA.bound_ops import *
from auto_LiRPA.utils import stop_criterion_batch_any, multi_spec_keep_func_all
from auto_LiRPA.utils import prod


class BaBSRNonlinearBranching:
    def __init__(self, net, num_branches):
        self.net = net
        self.model = net.net
        self.num_branches = num_branches

    def compute_heuristic(self, node, points, domains, margin_before):
        name = node.name
        num_neurons = np.prod(points.shape[1:-1])
        lAs = domains['lAs']
        lb, ub = domains['lower_bounds'], domains['upper_bounds']
        lb_ori, ub_ori = lb[name], ub[name]
        batch_size = lb_ori.shape[0]
        start_nodes = [act[0] for act in self.net.split_activations[name]]

        bounds_before = self.get_partial_bounds_batch(lb, ub, lAs, start_nodes)
        bounds_before = bounds_before.reshape(batch_size, -1)

        branched_neurons = torch.arange(
            num_neurons, device=bounds_before.device
        ).expand(batch_size, num_neurons)

        def _repeat(v):
            return v.repeat(num_neurons, *([1]*(v.ndim-1)))

        lAs = {k: _repeat(v) for k, v in lAs.items()
                if self.net.net[k] in start_nodes}
        lb = {k: _repeat(v) for k, v in lb.items()
            if k != name and k in self.interm_bound_required}
        ub = {k: _repeat(v) for k, v in ub.items()
            if k != name and k in self.interm_bound_required}
        margin_after = None
        for i in range(self.num_branches):
            lb_branched = lb_ori if i == 0 else points[..., i - 1]
            ub_branched = ub_ori if i == self.num_branches - 1 else points[..., i]
            bounds_after = self._get_bounds_after_branching(
                branched_neurons, lAs, lb, ub, lb_ori, ub_ori,
                lb_branched, ub_branched, name, start_nodes)
            margin_after_ = (
                margin_before
                + (bounds_after - bounds_before.unsqueeze(1))).amax(dim=-1)
            if i == 0:
                margin_after = margin_after_
            else:
                margin_after = torch.min(margin_after, margin_after_)
        return {
            'scores': margin_after - margin_before.amax(dim=-1),
            'points': points
        }

    def _get_bounds_after_branching(
            self, branched_neurons, lAs, lb, ub,
            lb_ori, ub_ori, lb_branched, ub_branched,
            node_name, start_nodes):
        batch_size, num_selected_neurons = branched_neurons.shape
        bounds_after = []

        def _branch(ori, branched, index):
            num_selected_neurons = index.shape[1]
            repeated = ori.repeat(
                num_selected_neurons, *([1]*(ori.ndim - 1))
            ).reshape(num_selected_neurons, *ori.shape)
            repeated = repeated.transpose(0, 1).reshape(
                batch_size, num_selected_neurons, -1)
            branched_ = torch.gather(branched.view(batch_size, -1),
                                     index=index, dim=-1)
            repeated.scatter_(dim=2, index=index.unsqueeze(-1),
                              src=branched_.unsqueeze(-1))
            return repeated.transpose(0, 1).reshape(-1, *branched.shape[1:])

        lAs_ = {k: v for k, v in lAs.items()}
        lb_ = {k: v for k, v in lb.items()}
        ub_ = {k: v for k, v in ub.items()}
        lb_[node_name] = _branch(lb_ori, lb_branched, branched_neurons)
        ub_[node_name] = _branch(ub_ori, ub_branched, branched_neurons)

        bounds_after.append(self.get_partial_bounds_batch(
            lb_, ub_, lAs_, start_nodes))
        bounds_after = torch.concat(bounds_after, dim=0)
        bounds_after = bounds_after.view(num_selected_neurons, batch_size, -1)
        bounds_after = bounds_after.transpose(0, 1)

        return bounds_after

    @torch.no_grad()
    def get_partial_bounds_batch(self, lb, ub, lAs, start_nodes):
        self.interm_bound_required = set()
        batch_size = next(iter(lb.values())).shape[0]
        x = self.net.expand_x(batch_size, lb=lb, ub=ub)
        interm_bounds = {k: (lb[k], ub[k]) for k in lb}
        self.model.set_input(x, interm_bounds=interm_bounds)
        for node in self.model.nodes():
            if not node.perturbed:
                fv = self.model.get_forward_value(node)
                node.interval = node.lower, node.upper = fv, fv

        for node in self.model.nodes():
            node.lA = node.uA = None
        start_nodes = list(set(start_nodes))
        degree_out = get_degrees(start_nodes)
        for node in start_nodes:
            node.lA = lAs[node.name].transpose(0, 1)
        for k in lb:
            self.model[k].lower = lb[k]
            self.model[k].upper = ub[k]
        lb = ub = torch.tensor(0., device=x.device)

        queue = deque(list(set(start_nodes)))
        while len(queue) > 0:
            node = queue.popleft()  # backward from l
            if node.name in self.model.root_names: continue

            # if all the succeeds are done, then we can turn to this node in the
            # next iteration.
            for node_pre in node.inputs:
                degree_out[node_pre.name] -= 1
                if degree_out[node_pre.name] == 0:
                    queue.append(node_pre)

            # Initially, l.lA or l.uA will be set to C for this node.
            if node.lA is not None or node.uA is not None:
                if not node.perturbed:
                    if not hasattr(node, 'forward_value'):
                        self.model.get_forward_value(node)
                    lb, ub = add_constant_node(lb, ub, node)
                    continue
                if node.zero_uA_mtx and node.zero_lA_mtx:
                    # A matrices are all zero, no need to propagate.
                    continue
                lb = lb + self._backward_propagate(node, start_nodes)

        assert lb.ndim == 2
        lb = lb.transpose(0, 1)

        return lb

    def _backward_propagate(self, node, start_nodes):
        for i in node.requires_input_bounds:
            self.interm_bound_required.add(node.inputs[i].name)
        A, lower_b, _ = node.bound_backward(
            node.lA, None, *node.inputs,
            start_node=self.model[self.model.final_name], start_shape=1)
        lb = lower_b

        if isinstance(lower_b, torch.Tensor) and node not in start_nodes:
            # For the BaBSR-like heuristic, stop further back
            # propagating, once we get some bias term from the layer
            # before the activation (typically a linear layer).
            return lb
        else:
            for i, node_pre in enumerate(node.inputs):
                add_bound(node, node_pre, lA=A[i][0])

        return lb



import json
import os
import time
import pickle
import subprocess
import traceback

import yaml

from CTRAIN.verification_systems.abCROWN.complete_verifier.abcrown import ABCROWN
from CTRAIN.verification_systems.abCROWN.complete_verifier.read_vnnlib import read_vnnlib
import torch
from CTRAIN.complete_verification.abCROWN.util import get_abcrown_standard_conf

MAX_LOSS = 10 ** 10

def limited_abcrown_eval(work_dir, runner_path='src/complete_verification/abCROWN/runner.py', *args, **kwargs):
    outer_timeout = kwargs['timeout'] * 1.2
    
    timestamp = time.time()
    
    args_pkl_path = f'{work_dir}/args_{timestamp}.pkl'
    result_path = f"{work_dir}/result_{timestamp}.pkl"
    
    with open(f'{work_dir}/args_{timestamp}.pkl', "wb") as f:
        pickle.dump((args, kwargs), f)
        
    verification_ok = False
    
    runner_args = [args_pkl_path, result_path]
    
    try:
        print(f"Running {['python3', runner_path] + runner_args}")
        process = subprocess.Popen(
            ["python3", runner_path] + runner_args,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        try:
            stdout, stderr = process.communicate(timeout=outer_timeout)
            print("Function finished successfully.")
            print("Output:", stdout.decode())
            print("Error Output:", stderr.decode())
            verification_ok = True

        except subprocess.TimeoutExpired:
            print(f"Function exceeded timeout of {outer_timeout} seconds. Terminating...")
            process.terminate()
            try:
                process.wait(timeout=5)
            except subprocess.TimeoutExpired:
                print("Function did not terminate after SIGTERM. Killing...")
                process.kill()

    except Exception as e:
        print(f"Error running the process: {e}")
    
    if verification_ok:
        with open(result_path, 'rb') as f:
            running_time, result = pickle.load(f)

        return running_time, result
    
    return MAX_LOSS, 'unknown'


def abcrown_eval(config, seed, instance, vnnlib_path='../../vnnlib/', model_name='mnist_6_100', model_path='./abCROWN/complete_verifier/models/eran/mnist_6_100_nat.pth', model_onnx_path=None, input_shape=[-1, 1, 28, 28], timeout=600, no_cores=28, par_factor=10):
    print(config, seed, instance)
    std_conf = config
    
    device = 'cuda' if torch.cuda.is_available() else 'mps' if torch.backends.mps.is_available() else 'cpu'
    
    timestamp = time.time()
    
    std_conf['model']['name'] = model_name
    std_conf['model']['path'] = f'/tmp/{model_name}.pth' if model_name is not None else None
    std_conf['model']['onnx_path'] = model_onnx_path if model_onnx_path is not None else None
    std_conf['model']['input_shape'] = input_shape
    
    std_conf['general']['device'] = device
    
    std_conf['bab']['timeout'] = timeout
    
    if not std_conf['solver'].get('mip'):
        std_conf['solver']['mip'] = get_abcrown_standard_conf(timeout=timeout, no_cores=no_cores)['solver']['mip']
    std_conf['solver']['mip']['parallel_solvers'] = no_cores
    
    std_conf['specification']['vnnlib_path_prefix'] = vnnlib_path
    std_conf['specification']['vnnlib_path'] = instance
    std_conf['general']['output_file'] = f'/tmp/out_{timestamp}.pkl'
        
    print(json.dumps(config, indent=2))
    
    with open(f"/tmp/conf_{timestamp}.yaml", "w", encoding='u8') as f:
        yaml.dump(std_conf, f)
    
    abcrown_instance = ABCROWN(
        ['--config', f'/tmp/conf_{timestamp}.yaml']
    )
    
    # Precompile VNN-LIB s.t. each run can access the cache
    _ = read_vnnlib(instance)
    
    start_time = time.time()
    try:
        verification_res = abcrown_instance.main()
    except Exception as e:
        print(type(e), e)
        print(traceback.format_exc())
        return MAX_LOSS, 'unknown'
    end_time = time.time()
    
    os.system(f'rm /tmp/conf_{timestamp}.yaml')
    
    with open(f'/tmp/out_{timestamp}.pkl', 'rb') as f:
        result_dict = pickle.load(f)
    
    result = result_dict['results']
    
    if result == 'unknown':
        print("PENALISING RUNNING TIME DUE TO TIMEOUT!")
        running_time = timeout * par_factor if timeout > (end_time - start_time) else (end_time - start_time) * par_factor
    else:
        running_time = end_time - start_time
    
    return running_time, result




from auto_LiRPA.perturbations import PerturbationLpNorm

import numpy as np
import torch

from CTRAIN.bound.ibp import bound_ibp
from CTRAIN.attacks import pgd_attack


def bound_sabr(hardened_model, original_model, data, target, eps, subselection_ratio, device='cuda', n_classes=10, x_L=None, x_U=None, data_min=None, data_max=None, n_steps=8, step_size=.5, restarts=1, early_stopping=True, intermediate_bound_model=None, decay_factor=0.1, decay_checkpoints=(4,7), return_adv_output=False):
    hardened_model.eval()
    original_model.eval()
    
    propagation_inputs, tau, x_adv = get_propagation_region(
        model=hardened_model,
        data=data,
        data_min=data_min,
        data_max=data_max,
        target=target,
        eps=eps if (x_L is None and x_U is None) else None,
        subselection_ratio=subselection_ratio,
        n_steps=n_steps,
        step_size=step_size,
        restarts=restarts,
        early_stopping=early_stopping,
        x_L=x_L,
        x_U=x_U,
        decay_checkpoints=decay_checkpoints, 
        decay_factor=decay_factor
    )
    
    hardened_model.train()
    original_model.train()
    
    ptb = PerturbationLpNorm(eps=tau, norm=np.inf, x_L=torch.clamp(propagation_inputs - tau, data_min, data_max).to(device), x_U=torch.clamp(propagation_inputs + tau, data_min, data_max).to(device))
    
    # Pass input through network to set batch statistics
    adv_output = hardened_model(x_adv)
    
    lb, ub = bound_ibp(
        model=hardened_model if intermediate_bound_model is None else intermediate_bound_model,
        ptb=ptb,
        data=propagation_inputs,
        target=target if intermediate_bound_model is None else None,
        n_classes=n_classes,
        bound_upper=True,
        reuse_input=False
    )
    
    if return_adv_output:
        return lb, ub, adv_output
    return lb, ub

def get_propagation_region(model, data, target, subselection_ratio, step_size, n_steps, restarts, x_L=None, x_U=None, data_min=None, data_max=None, eps=None, early_stopping=True, decay_factor=.1, decay_checkpoints=(4, 7)):
    device = torch.device("cuda" if torch.cuda.is_available() else "mps" if torch.backends.mps.is_available() else "cpu")
    assert (x_L is None and x_U is None and eps is not None) or (x_L is not None and x_U is not None and eps is None), "Please only provide epsilon value OR upper and lower input bounds"
    tau = None
    if eps.all() and data is not None:
        x_L=torch.clamp(data - eps, data_min, data_max).to(device)
        x_U=torch.clamp(data + eps, data_min, data_max).to(device)
    else:
        eps = torch.max((x_U - x_L))
    
    tau =  subselection_ratio * eps

    x_adv = pgd_attack(
        model=model,
        data=data,
        target=target,
        x_L=x_L,
        x_U=x_U,
        n_steps=n_steps,
        step_size=step_size,
        restarts=restarts,
        early_stopping=early_stopping,
        device=device,
        decay_checkpoints=decay_checkpoints,
        decay_factor=decay_factor
    )
    
    propagation_inputs = torch.clamp(x_adv, x_L + tau, x_U - tau) # called midpoints in SABR code
    tau = torch.tensor(tau, device=device)
    return propagation_inputs, tau, x_adv


#########################################################################
##   This file is part of the α,β-CROWN (alpha-beta-CROWN) verifier    ##
##                                                                     ##
##   Copyright (C) 2021-2024 The α,β-CROWN Team                        ##
##   Primary contacts: Huan Zhang <huan@huan-zhang.com>                ##
##                     Zhouxing Shi <zshi@cs.ucla.edu>                 ##
##                     Kaidi Xu <kx46@drexel.edu>                      ##
##                                                                     ##
##    See CONTRIBUTORS for all author contacts and affiliations.       ##
##                                                                     ##
##     This program is licensed under the BSD 3-Clause License,        ##
##        contained in the LICENCE file in this directory.             ##
##                                                                     ##
#########################################################################
import os
import copy
from collections import defaultdict

import torch
import arguments
import warnings

from auto_LiRPA import BoundedModule, BoundedTensor
from auto_LiRPA.perturbations import PerturbationLpNorm
from auto_LiRPA.utils import (
        stop_criterion_placeholder, stop_criterion_all, reduction_str2func)

from attack import attack_after_crown
from attack.attack_pgd import save_cex, process_vnn_lib_attack
from input_split.input_split_on_relu_domains import input_branching_decisions
from utils import Timer
from load_model import Customized
from prune import PruneAfterCROWN
from domain_updater import (DomainUpdater, DomainUpdaterSimple)
from heuristics.nonlinear import precompute_A


class LiRPANet:
    def __init__(self, model_ori, in_size, c=None, device=None,
                 cplex_processes=None):
        """
        convert pytorch model to auto_LiRPA module
        """
        general_args = arguments.Config['general']
        solver_args = arguments.Config['solver']
        model_args = arguments.Config['model']
        bab_args = arguments.Config['bab']
        self.c = c
        self.model_ori = model_ori
        self.input_shape = in_size
        self.device = device or general_args['device']
        model_ori_state_dict = copy.deepcopy(model_ori.state_dict())
        self.net = BoundedModule(
            model_ori, torch.zeros(in_size, device=self.device),
            bound_opts={
                'deterministic': general_args['deterministic'],
                'conv_mode': general_args['conv_mode'],
                'sparse_features_alpha': general_args['sparse_alpha'],
                'sparse_spec_alpha': general_args['sparse_alpha'],
                'sparse_intermediate_bounds': general_args['sparse_interm'],
                'crown_batch_size': solver_args['crown']['batch_size'],
                'max_crown_size': solver_args['crown']['max_crown_size'],
                'forward_refinement': solver_args['forward']['refine'],
                'dynamic_forward': solver_args['forward']['dynamic'],
                'forward_max_dim': solver_args['forward']['max_dim'],
                'use_full_conv_alpha': solver_args['alpha-crown']['full_conv_alpha'],
                'disable_optimization': solver_args['alpha-crown']['disable_optimization'],
                'fixed_reducemax_index': True,
                'matmul': {'share_alphas': solver_args['alpha-crown']['matmul_share_alphas']},
                'tanh': {'loose_threshold': bab_args['branching']['nonlinear_split']['loose_tanh_threshold']},
                'relu': solver_args['crown']['relu_option'],
                'buffers': {'has_batchdim': general_args['buffer_has_batchdim']},
                'optimize_bound_args': {
                    'apply_output_constraints_to': solver_args['invprop']['apply_output_constraints_to'],
                    'tighten_input_bounds': solver_args['invprop']['tighten_input_bounds'],
                    'best_of_oc_and_no_oc': solver_args['invprop']['best_of_oc_and_no_oc'],
                    'directly_optimize': solver_args['invprop']['directly_optimize'],
                    'oc_lr': solver_args['invprop']['oc_lr'],
                    'share_gammas': solver_args['invprop']['share_gammas'],
                },
                "optimize_graph": {
                    "optimizer": eval(model_args['optimize_graph']) if model_args['optimize_graph'] else None,
                }
            },
            device=self.device
        )
        self.net = eval(general_args['graph_optimizer'])(self.net)
        self.root = self.net[self.net.root_names[0]]

        self.net.eval()
        self.return_A = False
        self.needed_A_dict = None
        self.cutter = None # class for generating and optimizing cuts
        self.timer = Timer()

        # for fetching cplex in parallel
        self.mip_building_proc = self.processes = None
        self.cplex_processes = cplex_processes
        self.pool = self.pool_result = self.pool_termination_flag = None # For multi-process.

        # for recording whether we need to return intermediate bounds
        # after initial intermediate bounds fetching, this switch will be
        # aligned with arg.bab.interm_transfer
        self.interm_transfer = True

        self.final_name = self.net.final_name
        self.alpha_start_nodes = [self.final_name] + list(filter(
            lambda x: len(x.strip()) > 0,
            bab_args['optimized_interm'].split(',')))

        if arguments.Config['model']['with_jacobian']:
            print('Not checking the conversion correctness for this model with JacobianOP')
        else:
            # check conversion correctness
            dummy = torch.randn(in_size, device=self.device)
            try:
                assert torch.allclose(model_ori(dummy), self.net(dummy))
            except AssertionError:
                print('torch allclose failed: norm '
                    f'{torch.norm(model_ori(dummy) - self.net(dummy))}')

        model_ori.load_state_dict(model_ori_state_dict, strict=False)

    @property
    def split_nodes(self):
        return self.net.split_nodes

    @torch.no_grad()
    def get_primal_upper_bound(self, A):
        assert self.x.ptb.norm == torch.inf, (
            'We only support to get primals for Linf norm perturbation so far')
        input_A_lower = A[self.net.output_name[0]][self.net.input_name[0]]['lA']
        batch = input_A_lower.shape[0]

        x_lb, x_ub = self.x.ptb.x_L, self.x.ptb.x_U
        x_lb = x_lb.repeat(batch, 1, 1, 1)
        x_ub = x_ub.repeat(batch, 1, 1, 1)
        input_primal = x_lb.clone().detach()
        input_primal[input_A_lower.squeeze(1) < 0] = x_ub[input_A_lower.squeeze(1) < 0]

        assert self.c.size(0) == 1
        return input_primal, self.model_ori(input_primal).matmul(self.c[0].transpose(-1, -2))

    def get_interm_bounds(self, lb, ub=None, init=True, device=None):
        """Get the intermediate bounds.

        By default, we also add final layer bound after applying C
        (lb and lb+inf).
        """

        lower_bounds, upper_bounds = {}, {}
        if init:
            self._get_split_nodes()
            for layer in self.net.layers_requiring_bounds + self.net.split_nodes:
                lower_bounds[layer.name] = layer.lower.detach()
                upper_bounds[layer.name] = layer.upper.detach()
        elif self.interm_transfer:
            for layer in self.net.layers_requiring_bounds:
                lower_bounds[layer.name] = self._transfer(
                    layer.lower.detach(), device)
                upper_bounds[layer.name] = self._transfer(
                    layer.upper.detach(), device)

        lower_bounds[self.final_name] = lb.flatten(1).detach()
        if ub is None:
            ub = lb + torch.inf
        upper_bounds[self.final_name] = ub.flatten(1).detach()

        return lower_bounds, upper_bounds

    def get_mask(self):
        masks = {}
        for node in self.net.get_splittable_activations():
            mask = []
            for idx in node.requires_input_bounds:
                input_node = node.inputs[idx]
                if not input_node.perturbed:
                    mask.append(None)
                else:
                    mask.append(node.get_split_mask(
                        input_node.lower, input_node.upper, idx))
            masks[node.name] = mask
        return masks

    def get_lA(self, preserve_mask=None, tot_cells=None,
               transpose=True, device=None):
        lAs = {}
        nodes = list(self.net.get_splittable_activations())
        for node in nodes:
            lA = getattr(node, 'lA', None)
            if lA is None:
                continue
            if preserve_mask is not None:
                new_lA = torch.zeros(
                    [tot_cells, lA.shape[0]] + list(lA.shape[2:]),
                    dtype=lA.dtype, device=lA.device)
                new_lA[preserve_mask] = lA.transpose(0,1)
                lA = new_lA
            else:
                lA = lA.transpose(0, 1) if transpose else lA.squeeze(0)
            lAs[node.name] = self._transfer(lA, device)
        return lAs

    def get_candidate_parallel(self, lb, ub, device=None):
        """Get the intermediate bounds in the current model."""
        return self.get_interm_bounds(lb, ub, init=False, device=device)

    def expand_batch(self, x, batch):
        return x[0:1].expand(batch, *[-1]*(x.ndim-1))

    def expand_x(self, batch, x_L=None, x_U=None, lb=None, ub=None):
        if x_L is None and x_U is None:
            ptb = PerturbationLpNorm(
                norm=self.x.ptb.norm, eps=self.x.ptb.eps,
                x_L=self.expand_batch(self.x.ptb.x_L, batch),
                x_U=self.expand_batch(self.x.ptb.x_U, batch))
        else:
            ptb = PerturbationLpNorm(norm=self.x.ptb.norm, x_L=x_L, x_U=x_U)
        new_x = BoundedTensor(self.expand_batch(self.x.data, batch), ptb)
        return new_x

    @torch.no_grad()
    def _expand_tensors(self, d, batch):
        lb, ub = d['lower_bounds'], d['upper_bounds']
        cs, x_Ls, x_Us = d.get('cs', None), d.get('x_Ls', None), d.get('x_Us', None)
        # Only the last element is used later.
        lb_last, ub_last = lb[self.final_name], ub[self.final_name]
        interm_bounds = {k: [lb[k], ub[k]] for k in lb if k != self.final_name}
        # create new_x here since batch may change
        new_x = self.expand_x(batch, x_Ls, x_Us, lb, ub)
        if cs is None:
            assert self.c.size(0) == 1
            cs = None if self.c is None else self.c.expand(new_x.shape[0], -1, -1)
        return interm_bounds, lb_last, ub_last, cs, new_x, x_Ls, x_Us

    def update_bounds(self, d, beta=None, fix_interm_bounds=True,
                      shortcut=False, stop_criterion_func=stop_criterion_placeholder(),
                      multi_spec_keep_func=None, beta_bias=True):
        """Main function for computing bounds after branch and bound in Beta-CROWN."""
        deterministic_opt = arguments.Config['general']['deterministic_opt']
        solver_args = arguments.Config['solver']
        beta_args = solver_args['beta-crown']
        bab_args = arguments.Config['bab']
        if beta is None:
            # might need to set beta False in FSB node selection
            beta = beta_args['beta']
        vanilla_crown = bab_args['vanilla_crown']
        if vanilla_crown:
            alpha = beta = False
        else:
            alpha = True
        iteration = beta_args['iteration']
        get_upper_bound = bab_args['get_upper_bound']
        enable_opt_interm_bounds = beta_args['enable_opt_interm_bounds']
        branching_input_and_activation = bab_args['branching']['branching_input_and_activation']
        batch = d['upper_bounds'][self.final_name].shape[0]
        decision_thresh = d.get('thresholds', None)

        self.timer.start('func')
        self.timer.start('prepare')

        if self.net.cut_used:
            self.disable_cut_for_branching()
        if beta and not vanilla_crown:
            splits_per_example = self.set_beta(d, bias=beta_bias)
            self.net.cut_used = (
                arguments.Config['bab']['cut']['enabled']
                and arguments.Config['bab']['cut']['bab_cut']
                and getattr(self.net, 'cut_module', None) is not None)
            # even we need to use cut, maybe the cut is not fetched yet
            if self.net.cut_used:
                iteration = self.set_cut_params(
                    batch, batch, d.get('split_history', None))
            # here to handle the case where the split node happen to be in the
            # cut constraint !!!
        ret = self._expand_tensors(d, batch)
        interm_bounds, lb_last, ub_last, c, new_x, x_Ls, x_Us = ret

        if alpha:
            self.set_alpha(d, set_all=enable_opt_interm_bounds)
        self.timer.add('prepare')

        self.timer.start('bound')
        self.net.set_bound_opts({
            'optimize_bound_args': {
                'enable_beta_crown': beta,
                'fix_interm_bounds': fix_interm_bounds,
                'stop_criterion_func': stop_criterion_func,
                'multi_spec_keep_func': multi_spec_keep_func,
                'iteration': iteration,
            },
            'enable_opt_interm_bounds': enable_opt_interm_bounds,
        })
        self.set_crown_bound_opts('beta')

        if shortcut:
            args_compute_bounds = dict(x=(new_x,), C=c, reuse_alpha=True,
                interm_bounds=interm_bounds, bound_upper=False,
                decision_thresh=decision_thresh)
            with (torch.enable_grad() if beta else torch.no_grad()):
                lb, _, = self.net.compute_bounds(
                    method='CROWN-optimized' if beta else 'backward',
                    **args_compute_bounds)
            return lb

        # we need A matrix to construct adv example
        return_A = get_upper_bound or branching_input_and_activation
        original_size = new_x.shape[0]

        if fix_interm_bounds:
            reference_bounds = {}
            for name in self.alpha_start_nodes:
                if name in interm_bounds:
                    reference_bounds[name] = interm_bounds[name]
                    interm_bounds.pop(name)
        else:
            reference_bounds = interm_bounds
            interm_bounds = {}
        if len(reference_bounds):
            print('Recompute intermediate bounds for nodes:',
                ', '.join(list(reference_bounds.keys())))

        if vanilla_crown:
            method = 'CROWN'
        else:
            method = 'CROWN-optimized'
        tmp_ret = self.net.compute_bounds(
            x=(new_x,), C=c, method=method,
            interm_bounds=interm_bounds, reference_bounds=reference_bounds,
            return_A=return_A, needed_A_dict=self.needed_A_dict,
            cutter=self.cutter, bound_upper=False,
            decision_thresh=decision_thresh)
        self.timer.add('bound')

        if return_A:
            lb, _, A = tmp_ret
        else:
            lb, _ = tmp_ret
        if get_upper_bound:
            primal_x, ub = self.get_primal_upper_bound(A)
        else:
            ub = torch.full_like(lb, fill_value=torch.inf)  # dummy upper bound
            primal_x = None
        # Use A matrix of the input, the find best neuron to branch in input space.
        input_split_idx = input_branching_decisions(
            self.net, lb,
            A[self.net.output_name[0]][self.net.input_name[0]]['lA'],
            x_Ls, x_Us, decision_thresh
        ) if branching_input_and_activation else None

        with torch.no_grad():
            # Move tensors to CPU for all elements in this batch.
            self.timer.start('transfer')
            lb, ub = lb.to(device='cpu'), ub.to(device='cpu')
            lAs = self.get_lA(
                self.net.last_update_preserve_mask, original_size,
                device='cpu', transpose=True)
            self.timer.add('transfer')
            self.timer.start('finalize')
            if alpha:
                ret_s = self.get_alpha(device='cpu', half=not deterministic_opt)
            else:
                ret_s = {}
            if beta:
                ret_b = self.get_beta(splits_per_example, device='cpu')
            else:
                ret_b = [{} for _ in range(batch)]
            # Reorganize tensors.
            ret_l, ret_u = self.get_candidate_parallel(lb, ub, device='cpu')
            if not deterministic_opt:
                ret_l[self.final_name] = torch.max(
                    ret_l[self.final_name], lb_last.cpu())
                if not get_upper_bound:
                    # Do not set to min so the primal is always corresponding
                    # to the upper bound.
                    ret_u[self.final_name] = torch.min(
                        ret_u[self.final_name], ub_last.cpu())
            self.timer.add('finalize')

        # Each key is corresponding to a pre-relu layer, and each value intermediate
        # beta values for neurons in that layer.
        new_split_history = [{} for _ in range(batch)]
        if self.net.cut_used:
            self.set_cut_new_split_history(new_split_history, batch)

        self.timer.add('func')
        self.timer.print()

        # Each key is corresponding to a pre-relu layer, and each value
        # intermediate beta values for neurons in that layer.,
        best_intermediate_betas = [defaultdict(dict) for _ in range(batch)]

        return {
            'lower_bounds': ret_l, 'upper_bounds': ret_u,
            'lAs': lAs, 'alphas': ret_s,
            'betas': ret_b, 'split_history': new_split_history,
            'intermediate_betas': best_intermediate_betas,
            'primals': primal_x,
            'c': c, 'x_Ls': x_Ls, 'x_Us': x_Us,
            'input_split_idx': input_split_idx,
        }

    def _get_split_nodes(self, verbose=False):
        self.net.get_split_nodes()
        self.split_activations = self.net.split_activations
        if verbose:
            print('Split layers:')
            for layer in self.net.split_nodes:
                print(f'  {layer}: {self.split_activations[layer.name]}')
            print('Nonlinear functions:')
            for node in self.net.nodes():
                if node.perturbed and len(node.requires_input_bounds):
                    print('  ', node)

    def build(self, input_domain, x,
              stop_criterion_func=stop_criterion_placeholder(),
              bounding_method=None, decision_thresh=None, vnnlib_ori=None,
              interm_bounds=None):
        # TODO merge with build_with_refined_bounds()
        solver_args = arguments.Config['solver']
        bab_args = arguments.Config['bab']
        branching_args = bab_args['branching']
        share_alphas = solver_args['alpha-crown']['share_alphas']
        bounding_method = bounding_method or solver_args['bound_prop_method']
        branching_input_and_activation = branching_args['branching_input_and_activation']
        fast_A = (branching_args['method'] == 'nonlinear'
                  and branching_args['nonlinear_split']['method'] == 'shortcut')

        self.x = x
        self.input_domain = input_domain
        self.net.set_bound_opts({
            'optimize_bound_args': {'stop_criterion_func': stop_criterion_func},
            'verbosity': 1,
        })
        self.set_crown_bound_opts('alpha')
        self._get_split_nodes(verbose=True)

        # expand x to align with C's batch size for multi target verification
        batch = self.c.size()[0]
        if self.x.shape[0] == 1 and batch > 1:
            x_expand = self.expand_x(batch)
        else:
            x_expand = x

        result = {key: None for key in [
            'mask', 'lA', 'lower_bounds', 'upper_bounds',
            'alphas', 'history', 'input_split_idx', 'attack_images']}

        self._set_A_options()
        prune_after_crown = None
        if bounding_method == 'alpha-crown':
            # first get CROWN bounds
            # Reference bounds are intermediate layer bounds from initial CROWN bounds.
            lb, ub, aux_reference_bounds = self.net.init_alpha(
                (x_expand,), share_alphas=share_alphas, c=self.c, bound_upper=False)
            print('initial CROWN bounds:', lb, ub)

            if arguments.Config['general']['save_output']:
                arguments.Globals['out']['init_crown_bounds'] = lb.cpu()

            if bab_args['cut']['enabled'] or bab_args['cut']['implication']:
                self.enable_cuts()

            if stop_criterion_func(lb).all().item():
                # Fast path. Initial CROWN bound can verify the network.
                print('Verified with initial CROWN!')
                return lb, result

            if arguments.Config['attack']['pgd_order'] == 'middle' and vnnlib_ori is not None:
                # Run adversarial attack on those specs that cannot be verified by CROWN.
                verified_success, attack_images = attack_after_crown(
                    lb, vnnlib_ori[0], self.model_ori, x, decision_thresh)
                if verified_success:  # Adversarial images are generated here.
                    print('PGD attack succeeded.')
                    if arguments.Config["general"]["save_adv_example"]:
                        try:
                            attack_output = self.model_ori(attack_images.view(-1, *attack_images.shape[1:]))
                            list_target_label_arrays, data_min_repeat, data_max_repeat = process_vnn_lib_attack(vnnlib_ori, x)
                            save_cex(attack_images, attack_output, x, vnnlib_ori, arguments.Config["attack"]["cex_path"], data_max_repeat, data_min_repeat)
                        except Exception as e:
                            print(str(e))
                            print('save adv example failed')
                    result.update({'attack_images': attack_images})
                    return lb, result

            c_to_use = self.c
            if solver_args['prune_after_crown']:
                prune_after_crown = PruneAfterCROWN(
                    self.net, self.c, lb,
                    decision_thresh=decision_thresh)
                c_to_use = prune_after_crown.c
                # This should be the only supported case for incomplete verifier.
                assert stop_criterion_func.__qualname__.split('.')[0] == 'stop_criterion_all'
                pruned_stop_criterion_func = stop_criterion_all(prune_after_crown.decision_thresh)
                self.net.set_bound_opts({
                    'optimize_bound_args': {'stop_criterion_func': pruned_stop_criterion_func},
                })

            ret = self.net.compute_bounds(
                x=(x_expand,), C=c_to_use, method='CROWN-Optimized',
                return_A=self.return_A, needed_A_dict=self.needed_A_dict,
                bound_upper=False, aux_reference_bounds=aux_reference_bounds,
                cutter=self.cutter, interm_bounds=interm_bounds)
        elif bounding_method == 'alpha-forward':
            warnings.warn('alpha-forward can only be used with input split for now')
            self.net.bound_opts['optimize_bound_args']['init_alpha'] = True
            ret = self.net.compute_bounds(
                x=(x_expand,), C=self.c, method='forward-optimized', bound_upper=False)
        else:
            with torch.no_grad():
                if bounding_method == 'init-crown':
                    assert not self.return_A
                    lb, ub, _ = self.net.init_alpha(
                        (x_expand,), share_alphas=share_alphas, c=self.c,
                        bound_upper=False)
                    ret = lb, ub
                else:
                    ret = self.net.compute_bounds(
                        x=(x_expand,), C=self.c, method=bounding_method,
                        bound_upper=False, return_A=self.return_A,
                        needed_A_dict=self.needed_A_dict)

        if self.return_A:
            lb, _, A = ret
        else:
            lb, _ = ret
            A = None

        if branching_input_and_activation:
            # Use A matrix of the input, the find best neuron to branch in input space.
            input_split_idx = input_branching_decisions(
                self.net, lb,
                A[self.net.output_name[0]][self.net.input_name[0]]['lA'],
                x_expand.ptb.x_L, x_expand.ptb.x_U, decision_thresh)
        else:
            input_split_idx = None
        if prune_after_crown:
            lb = prune_after_crown.recover_lb(lb)

        print(f'initial {bounding_method} bounds:', lb)

        # save initial alpha-crown for tests
        if arguments.Config['general']['save_output'] and bounding_method == 'alpha-crown':
            arguments.Globals['out']['init_alpha_crown'] = lb.cpu()

        global_lb = lb.min().item()
        print('Worst class: (+ rhs)', global_lb)

        # DEBUG: check loose bounds
        if os.environ.get('ABCROWN_VIEW_INTERM', False):
            print('Intermediate bounds after initial alpha-CROWN:')
            self._print_interm_bounds()
            import pdb; pdb.set_trace()

        alpha = self.get_alpha()  # initial with one node only
        # for each pre-activation layer, we initial 2 lists for the two branches
        lb, ub = self.get_interm_bounds(lb)  # primals are better upper bounds
        history = self.empty_history()
        mask = self.get_mask()
        lA = self.get_lA()
        if prune_after_crown:
            lA = prune_after_crown.recover_lA_and_alpha(lA, alpha)

        if fast_A and global_lb < 0 and self.return_A:
            precompute_A(self.net, A, x_expand,
                         interm_bounds={k: (lb[k], ub[k]) for k in lb})
        if bab_args['cut']['enabled']:
            self.create_cutter(A, x_expand, lb, ub, prune_after_crown)
        if A is not None or bab_args['cut']['implication']:
            # FIXME There is not only A here. There are also biases.
            # Need to rename.
            self.A_saved = A

        result.update({
            'mask': mask, 'lA': lA, 'lower_bounds': lb, 'upper_bounds': ub,
            'alphas': alpha, 'history': history,
            'input_split_idx': input_split_idx,
        })
        return lb[self.final_name], result

    def build_with_refined_bounds(
            self, input_domain, x,
            refined_lower_bounds=None, refined_upper_bounds=None,
            activation_opt_params=None, reference_lA=None,
            reference_alphas=None, cutter=None, refined_betas=None,
            stop_criterion_func=stop_criterion_placeholder(),
            decision_thresh=None):
        solver_args = arguments.Config['solver']
        bab_args = arguments.Config['bab']
        branch_args = bab_args['branching']
        share_alphas = solver_args['alpha-crown']['share_alphas']
        target_batch_size = solver_args['multi_class']['label_batch_size']
        loss_reduction_func = arguments.Config['general']['loss_reduction_func']
        branching_input_and_activation = branch_args['branching_input_and_activation']
        vanilla_crown = bab_args['vanilla_crown']

        self.x = x
        self.input_domain = input_domain
        self.cutter = cutter

        # expand x to align with C's batch size for multi target verification
        batch = self.c.size()[0]
        x_expand = self.expand_x(batch)
        # also, we need to expand lower and upper bounds accordingly
        if refined_lower_bounds is not None and refined_upper_bounds is not None:
            # the intermediate bounds were shared in incomplete_verifier(), we expand them here
            for k, v in refined_lower_bounds.items():
                if k != self.final_name:
                    refined_lower_bounds[k] = self.expand_batch(v, batch)
            for k, v in refined_upper_bounds.items():
                if k != self.final_name:
                    refined_upper_bounds[k] = self.expand_batch(v, batch)

        loss_reduction_func = reduction_str2func(loss_reduction_func)
        self.refined_lower_bounds = refined_lower_bounds
        self.refined_upper_bounds = refined_upper_bounds

        self._set_A_options(bab=True)

        # batch results holder
        batch_lbs, batch_ubs, lA, alphas = [], [], {}, {}
        # tot label batches
        tot_batches = (x_expand.size()[0] + target_batch_size - 1) // target_batch_size

        def _take_batch(x, idx):
            return x[idx * target_batch_size: (idx + 1) * target_batch_size]

        for now_batch in range(tot_batches):
            print(f'build_with_refined_bounds batch [{now_batch+1}/{tot_batches}]')
            torch.cuda.empty_cache()

            batch_expand = BoundedTensor(
                _take_batch(x_expand.data, now_batch),
                PerturbationLpNorm(
                    x_expand.ptb.eps, x_expand.ptb.norm,
                    _take_batch(x_expand.ptb.x_L, now_batch),
                    _take_batch(x_expand.ptb.x_U, now_batch)))
            C_batch = _take_batch(self.c, now_batch)

            # For updating which nodes are perturbed.
            self.net.set_input(batch_expand)
            if (refined_lower_bounds is not None
                    and refined_upper_bounds is not None):
                # using refined bounds with init opt crown
                interm_bounds = {
                    k: [_take_batch(refined_lower_bounds[k], now_batch),
                        _take_batch(refined_upper_bounds[k], now_batch)]
                    for k in refined_lower_bounds
                }
            else:
                interm_bounds = None

            skip_backward_pass = False
            if vanilla_crown:
                ret = self.net.compute_bounds(
                    x=(batch_expand,), method='backward', C=C_batch,
                    return_A=self.return_A, #reuse_alpha=True,
                    interm_bounds=interm_bounds,
                    needed_A_dict=self.needed_A_dict)
            else:
                self.net.init_alpha(
                    (batch_expand,), share_alphas=share_alphas, c=C_batch,
                    interm_bounds=interm_bounds,
                    activation_opt_params=activation_opt_params,
                    skip_bound_compute=True)

                all_alpha_initialized = self.copy_alpha(
                    reference_alphas, batch_size=batch_expand.shape[0],
                    num_targets=min((now_batch + 1) * target_batch_size,
                                    self.c.shape[0]) - now_batch * target_batch_size,
                    target_batch_size=target_batch_size, now_batch=now_batch,
                    interm_bounds=interm_bounds)

                self.net.set_bound_opts({'optimize_bound_args': {
                    'stop_criterion_func': stop_criterion_func,
                }})
                self.set_crown_bound_opts('alpha')

                if all_alpha_initialized and solver_args['multi_class']['skip_with_refined_bound']:
                    print('all alpha initialized')
                    if not self.return_A:
                        # FIXME "A" is incorrect later when calling get_lA
                        skip_backward_pass = True
                        print('directly get lb and ub from refined bounds')
                        # Make sure the shape of reference_lA looks good so that we
                        # can recover the batch_lA
                        print('c shape:', self.c.shape)
                        print('lA shapes:', [A.shape for A in reference_lA.values()])
                        # A shape: [batch, num_output, *output_shape]
                        assert all([A.shape[0] == self.c.shape[0] for A in reference_lA.values()])
                        # Try to directly recover l and u from refined_lower_bounds
                        # and refined_upper_bounds without a backward crown pass
                        # refined_lower/upper_bounds[-1]'s shape is [labels to verify, C]
                        # self.c's shape is [labels to verify, 1, C] where target labels have value -1.
                        lb = _take_batch(refined_lower_bounds[self.final_name], now_batch)
                        ub = _take_batch(refined_upper_bounds[self.final_name], now_batch)
                        ret = (lb, ub)
                    else:
                        # do a backward crown pass
                        print('true A is required, we do a full backward CROWN pass to obtain it')
                        ret = self.net.compute_bounds(
                            x=(batch_expand,), method='backward', C=C_batch,
                            return_A=self.return_A, reuse_alpha=True,
                            interm_bounds=interm_bounds,
                            needed_A_dict=self.needed_A_dict)
                else:
                    print('Restore to original setting since some alphas are not '
                        'initialized yet or being asked not to skip')
                    ret = self.net.compute_bounds(
                        x=(batch_expand,), method='crown-optimized',
                        return_A=self.return_A, C=C_batch,
                        interm_bounds=interm_bounds,
                        needed_A_dict=self.needed_A_dict)

            if self.return_A:
                lb, ub, A = ret
            else:
                lb, ub = ret
                A = None

            if branching_input_and_activation:
                # Use A matrix of the input, the find best neuron to branch in input space.
                input_split_idx = input_branching_decisions(
                    self.net, lb,
                    A[self.net.output_name[0]][self.net.input_name[0]]['lA'],
                    x.ptb.x_L, x.ptb.x_U, decision_thresh)
            else:
                input_split_idx = None

            batch_lb, batch_ub = self.get_interm_bounds(lb)  # primals are better upper bounds

            print('(alpha-)CROWN with fixed intermediate bounds:', lb, ub)
            print('Intermediate layers:', ','.join(list(interm_bounds.keys())))
            if vanilla_crown:
                history = ret_b = None
            else:
                self.add_batch_alpha(alphas, reference_alphas)
                if refined_betas is not None:
                    # only has batch size 1 for refined betas
                    assert len(refined_betas[0]) == 1
                    history, ret_b = refined_betas[0][0], refined_betas[1][0]
                else:
                    history, ret_b = self.empty_history(), None

            mask = self.get_mask()
            if skip_backward_pass:
                # reference_lA is already transposed back in incomplete_verifier()
                batch_lA = reference_lA
            else:
                batch_lA = self.get_lA()

            batch_lbs.append(batch_lb)
            batch_ubs.append(batch_ub)
            for k, v in batch_lA.items():
                if k not in lA:
                    lA[k] = v
                else:
                    # Need to accumulate itemwise over the 0 dim, since A's shape
                    # is [batch, spec=1, ...]
                    lA[k] = torch.cat([lA[k], v], dim=0)

        # merge all things from the batch
        lb = {k: torch.cat([item_lb[k] for item_lb in batch_lbs])
              for k in batch_lbs[0]}
        ub = {k: torch.cat([item_ub[k] for item_ub in batch_ubs])
              for k in batch_ubs[0]}

        return {
            'global_ub': ub[self.final_name], 'global_lb': lb[self.final_name],
            'mask': mask, 'lA': lA, 'lower_bounds': lb, 'upper_bounds': ub,
            'alphas': alphas, 'history': history, 'betas': ret_b,
            'input_split_idx': input_split_idx,
        }

    def build_history_and_set_bounds(self, d, split, mode='depth', impl_params=None):
        _, num_split = DomainUpdater.get_num_domain_and_split(
            d, split, self.final_name)
        args = (self.root, self.final_name, self.net.split_nodes)
        if num_split == 1 and (split.get('points', None) is None
                                 or split['points'].ndim == 1):
            domain_updater = DomainUpdaterSimple(*args)
        else:
            domain_updater = DomainUpdater(*args)
        domain_updater.set_branched_bounds(d, split, mode)

    def _set_A_options(self, bab=False):
        branching_args = arguments.Config['bab']['branching']
        input_and_act = branching_args['branching_input_and_activation']
        fast_A = (branching_args['nonlinear_split']['method'] == 'shortcut'
                and branching_args['method'] == 'nonlinear')
        get_upper_bound = bab and arguments.Config['bab']['get_upper_bound']
        if get_upper_bound or input_and_act:
            self.needed_A_dict = defaultdict(set)
            self.needed_A_dict[self.net.output_name[0]].add(
                self.net.input_name[0])
        if fast_A:
            self.needed_A_dict = defaultdict(set)
            for node in self.net.nodes():
                if node != self.net.final_name:
                    self.needed_A_dict[node.name].add(self.net.input_name[0])
        # FIXME just use "self.needed_A_dict is not None" without the extra "self.return_A"
        if self.needed_A_dict is not None:
            self.return_A = True

    def empty_history(self):
        return {layer.name: ([], [], []) for layer in self.net.split_nodes}

    def _transfer(self, tensor, device=None, half=False):
        if half:
            tensor = tensor.half()
        if device:
            tensor = tensor.to(device)
        return tensor

    def set_crown_bound_opts(self, crown_name):
        solver_args = arguments.Config['solver']
        bab_args = arguments.Config['bab']
        crown_args = solver_args[f'{crown_name}-crown']
        opt_bound_args = {
            'deterministic': arguments.Config['general']['deterministic_opt'],
            'lr_alpha': crown_args['lr_alpha'],
            'iteration': crown_args['iteration'],
            'lr_decay': crown_args['lr_decay'],
            'use_float64_in_last_iteration': solver_args['use_float64_in_last_iteration'],
            'start_save_best': solver_args['start_save_best'],
            'loss_reduction_func': reduction_str2func(
                arguments.Config['general']['loss_reduction_func']),
        }
        if crown_name == 'alpha':
            opt_bound_args.update({
                'enable_alpha_crown': True, 'enable_beta_crown': False,
                'init_alpha': False, 'fix_interm_bounds': True,
                'use_shared_alpha': crown_args['share_alphas'],
                'early_stop_patience': solver_args['early_stop_patience'],
                'pruning_in_iteration': False
            })
        elif crown_name == 'beta':
            opt_bound_args.update({
                'optimizer': crown_args['optimizer'],
                'lr_beta': crown_args['lr_beta'],
                'pruning_in_iteration': bab_args['pruning_in_iteration'],
                'pruning_in_iteration_threshold': bab_args['pruning_in_iteration_ratio'],
                'lr_cut_beta': bab_args['cut']['lr_beta'],
                'apply_output_constraints_to': [],
                'tighten_input_bounds': False,
                'directly_optimize': [],
                'share_gammas': False,
            })
        self.net.set_bound_opts({'optimize_bound_args': opt_bound_args})

    def _print_interm_bounds(self, lb=None, ub=None):
        if lb is not None:
            for k in lb:
                print(self.net[k])
                for out in self.net[k].output_name:
                    print('  ', self.net[out])
                print('  lower:', lb[k])
                if ub is not None and k in ub:
                    print('  upper:', ub[k])
                unstable = torch.logical_and(lb[k] < 0, ub[k] > 0).float()
                unstable = unstable.reshape(unstable.size(0), -1).sum(dim=-1)
                print('  unstable:', unstable, unstable.mean(), unstable.max())
        else:
            for node in self.net.nodes():
                if not node.perturbed:
                    continue
                if isinstance(node.lower, torch.Tensor):
                    print(node)
                    print('  lower:', node.lower.reshape(-1)[:10])
                    if isinstance(node.upper, torch.Tensor):
                        print('  upper:', node.upper.reshape(-1)[:10])
                        print(' Average gap:', (node.upper-node.lower).mean())

    from alpha import get_alpha, set_alpha, copy_alpha, add_batch_alpha
    from beta import get_beta, set_beta, reset_beta
    from lp_mip_solver import (
        build_solver_model, update_mip_model_fix_relu,
        build_the_model_mip_refine, build_the_model_lp, build_the_model_mip,
        all_node_split_LP, check_lp_cut, update_the_model_cut)
    from input_split.bounding import get_lower_bound_naive
    from cuts.cut_verification import (
        enable_cuts, create_cutter, set_cuts, create_mip_building_proc,
        set_cut_params, set_cut_new_split_history,
        disable_cut_for_branching, set_dependencies)
    from prune import prune_reference_alphas, prune_lA



#########################################################################
##   This file is part of the α,β-CROWN (alpha-beta-CROWN) verifier    ##
##                                                                     ##
##   Copyright (C) 2021-2024 The α,β-CROWN Team                        ##
##   Primary contacts: Huan Zhang <huan@huan-zhang.com>                ##
##                     Zhouxing Shi <zshi@cs.ucla.edu>                 ##
##                     Kaidi Xu <kx46@drexel.edu>                      ##
##                                                                     ##
##    See CONTRIBUTORS for all author contacts and affiliations.       ##
##                                                                     ##
##     This program is licensed under the BSD 3-Clause License,        ##
##        contained in the LICENCE file in this directory.             ##
##                                                                     ##
#########################################################################
"""Branch and bound for input space split."""

import time
import torch
import math
import sys

import arguments
from auto_LiRPA.utils import stop_criterion_batch_any, stop_criterion_all
from input_split.branching_domains import UnsortedInputDomainList
from input_split.attack import (massive_pgd_attack, check_adv,
                                attack_in_input_bab_parallel,
                                update_rhs_with_attack)
from input_split.branching_heuristics import input_split_branching
from input_split.split import input_split_parallel, get_split_depth

Visited, Solve_alpha, storage_depth = 0, False, 0


def batch_verification_input_split(
        d, net, batch, num_iter, decision_thresh, shape=None,
        bounding_method="crown", branching_method="sb",
        stop_func=stop_criterion_batch_any, split_partitions=2):
    input_split_args = arguments.Config["bab"]["branching"]["input_split"]

    split_start_time = time.time()
    global Visited

    # STEP 1: find the neuron to split and create new split domains.
    pickout_start_time = time.time()
    ret = d.pick_out_batch(batch, device=net.x.device)
    alphas, dm_lb, x_L, x_U, cs, thresholds, split_idx, last_split_idx = ret
    pickout_time = time.time() - pickout_start_time

    if input_split_args['update_rhs_with_attack']:
        thresholds = update_rhs_with_attack(x_L, x_U, cs, thresholds, dm_lb,
                                            net.model_ori)

    # STEP 2: find the neuron to split and create new split domains.
    decision_start_time = time.time()
    split_depth = get_split_depth(x_L, split_partitions=split_partitions)
    new_x_L, new_x_U, cs, thresholds, split_depth, last_split_idx = input_split_parallel(
        x_L, x_U, shape, cs, thresholds, split_depth=split_depth, i_idx=split_idx,
        split_partitions=split_partitions)

    if input_split_args["compare_with_old_bounds"]:
        assert split_depth == 1
        dm_lb = dm_lb.repeat(2, *[1]*(dm_lb.ndim - 1))
    else:
        dm_lb = None

    alphas = alphas * (split_partitions ** (split_depth - 1))

    decision_time = time.time() - decision_start_time

    # STEP 3: Compute bounds for all domains.
    bounding_start_time = time.time()
    ret = net.get_lower_bound_naive(
        dm_lb=dm_lb, dm_l=new_x_L, dm_u=new_x_U, alphas=alphas,
        bounding_method=bounding_method, branching_method=branching_method,
        C=cs, stop_criterion_func=stop_func, thresholds=thresholds,
        num_iter=num_iter, split_partitions=split_partitions)

    # here alphas is a dict
    new_dm_lb, alphas, lA = ret
    bounding_time = time.time() - bounding_start_time

    decision_time -= time.time()
    split_idx = input_split_branching(
        net, new_dm_lb, new_x_L, new_x_U, lA, thresholds,
        branching_method, storage_depth, num_iter=num_iter,
        last_split_idx=last_split_idx)

    decision_time += time.time()
    # STEP 4: Add new domains back to domain list.
    adddomain_start_time = time.time()
    d.add(new_dm_lb, new_x_L.detach(), new_x_U.detach(),
          alphas, cs, thresholds, split_idx, last_split_idx=last_split_idx)
    adddomain_time = time.time() - adddomain_start_time

    total_time = time.time() - split_start_time
    print(
        f"Total time: {total_time:.4f}  pickout: {pickout_time:.4f} "
        f"decision: {decision_time:.4f}  bounding: {bounding_time:.4f} "
        f"add_domain: {adddomain_time:.4f}"
    )
    print("Length of domains:", len(d))

    Visited += len(new_x_L)
    print(f"{Visited} branch and bound domains visited")

    if len(d) == 0:
        print("No domains left, verification finished!")
        if new_dm_lb is not None:
            print(f"The lower bound of last batch is {new_dm_lb.min().item()}")
        return decision_thresh.max() + 1e-7
    else:
        worst_idx = d.get_topk_indices().item()
        worst_val = d[worst_idx]
        global_lb = worst_val[0] - worst_val[-1]

    if 1 < global_lb.numel() <= 5:
        print(f'Current (lb-rhs): {global_lb}')
    else:
        print(f'Current (lb-rhs): {global_lb.max().item()}')
    if input_split_args['show_progress']:
        print(f'Progress: {d.get_progess():.10f}')
    sys.stdout.flush()

    return global_lb


def input_bab_parallel(net, init_domain, x, rhs=None,
                       timeout=None, max_iterations=None,
                       vnnlib=None, c_transposed=False, return_domains=False):
    """Run input split bab.

    c_transposed: bool, by default False, indicating whether net.c matrix has
        transposed between dim=0 and dim=1. As documented in abcrown.py bab(),
        if using input split, and if there are multiple specs with shared single input,
        we transposed the c matrix from [multi-spec, 1, ...] to [1, multi-spec, ...] so that
        net.build() process in input_bab_parallel could share intermediate layer
        bounds across all specs. If such transpose happens, c_transposed is set
        to True, so that after net.build() in this func, we can transpose c
        matrix back, repeat x_LB & x_UB, duplicate alphas, to prepare for input domain bab.
    """
    global storage_depth

    start = time.time()
    # All supported arguments.
    global Visited, all_node_split

    bab_args = arguments.Config["bab"]
    branching_args = bab_args["branching"]
    input_split_args = branching_args["input_split"]

    timeout = timeout or bab_args["timeout"]
    batch = arguments.Config["solver"]["batch_size"]
    auto_enlarge_batch_size = arguments.Config["solver"]["auto_enlarge_batch_size"]
    bounding_method = arguments.Config["solver"]["bound_prop_method"]
    init_bounding_method = arguments.Config["solver"]["init_bound_prop_method"]
    max_iterations = max_iterations or bab_args['max_iterations']
    sort_domain_iter = bab_args["sort_domain_interval"]
    branching_method = branching_args['method']
    adv_check = input_split_args['adv_check']
    split_partitions = input_split_args['split_partitions']
    catch_assertion = input_split_args['catch_assertion']

    if net.device != 'cpu' and auto_enlarge_batch_size:
        total_vram = torch.cuda.get_device_properties(net.device).total_memory

    if init_bounding_method == "same":
        init_bounding_method = bounding_method

    if c_transposed or net.c.shape[1] == 1:
        # When c_transposed applied, previous checks have ensured that there is only single spec,
        # but we compressed multiple data samples to the spec dimension by transposing C
        # so we need "all" to be satisfied to stop.
        # Another case is there is only single spec, so batch_any equals to all. Then, we set to all
        # so we can allow prune_after_crown optimization
        stop_func = stop_criterion_all
    else:
        # Possibly multiple specs in each data sample
        stop_func = stop_criterion_batch_any

    Visited = 0

    dm_l = x.ptb.x_L
    dm_u = x.ptb.x_U

    if (dm_u - dm_l > 0).int().sum() == 1:
        branching_method = 'naive'

    try:
        global_lb, ret = net.build(
            init_domain, x, stop_criterion_func=stop_func(rhs),
            bounding_method=init_bounding_method, decision_thresh=rhs)
    except AssertionError:
        if catch_assertion:
            global_lb = torch.ones(net.c.shape[0], net.c.shape[1],
                                   device=net.device) * torch.inf
            ret = {'alphas': {}}
        else:
            raise

    if getattr(net.net[net.net.input_name[0]], 'lA', None) is not None:
        lA = net.net[net.net.input_name[0]].lA.transpose(0, 1)
    else:
        lA = None
        if bounding_method == 'sb':
            raise ValueError("sb heuristic cannot be used without lA.")

    if c_transposed:
        lA, global_lb, rhs, dm_l, dm_u = transpose_c_back(
            lA, global_lb, rhs, dm_l, dm_u, ret, net)

    result = "unknown"

    # compute storage depth
    use_alpha = init_bounding_method.lower() == 'alpha-crown' or bounding_method == 'alpha-crown'
    min_batch_size = (
        arguments.Config["solver"]["min_batch_size_ratio"]
        * arguments.Config["solver"]["batch_size"]
    )
    max_depth = max(int(math.log(max(min_batch_size, 1)) // math.log(split_partitions)), 1)
    storage_depth = min(max_depth, dm_l.shape[-1])
    domains = UnsortedInputDomainList(
        storage_depth, use_alpha=use_alpha,
        sort_index=input_split_args['sort_index'],
        sort_descending=input_split_args['sort_descending'])

    initial_verified, remaining_index = initial_verify_criterion(global_lb, rhs)
    if initial_verified:
        result = "safe"
    else:
        # compute initial split idx
        split_idx = input_split_branching(
            net, global_lb, dm_l, dm_u, lA, rhs, branching_method, storage_depth)
        domains.add(global_lb, dm_l.detach(), dm_u.detach(),
                    ret['alphas'], net.c, rhs, split_idx, remaining_index)
        if arguments.Config["attack"]["pgd_order"] == "after":
            if attack_in_input_bab_parallel(net.model_ori, domains, x, vnnlib=vnnlib).all():
                print("pgd attack succeed in input_bab_parallel")
                result = "unsafe"

    num_iter = 1
    enhanced_bound_initialized = False
    while (result == "unknown" and len(domains) > 0
           and (max_iterations == -1 or num_iter <= max_iterations)):
        print(f'Iteration {num_iter}')
        # sort the domains every certain number of iterations
        if sort_domain_iter > 0 and num_iter % sort_domain_iter == 0:
            sort_start_time = time.time()
            domains.sort()
            sort_time = time.time() - sort_start_time
            print(f"Sorting domains used {sort_time:.4f}s")

        last_glb = global_lb.max()

        if arguments.Config["attack"]["pgd_order"] != "skip":
            if adv_check != -1 and Visited > adv_check:
                adv_check_start_time = time.time()
                # check whether adv example found
                if check_adv(domains, net.model_ori, vnnlib=vnnlib):
                    return global_lb.max(), Visited, "unsafe"
                adv_check_time = time.time() - adv_check_start_time
                print(f"Adv attack time: {adv_check_time:.4f}s")

        if net.device is not 'cpu' and auto_enlarge_batch_size:
            current_vram = torch.cuda.memory_reserved()
            if current_vram < 0.45 * total_vram and batch < len(domains) and num_iter > 1:
                if batch * 2 > len(domains):
                    # in case the real batch_size = len(domains) in the last run
                    # we can at most set batch = len(domain)
                    batch = len(domains)
                else:
                    batch *= 2
                print('current_vram/total_varm: '
                      f'{current_vram/1e9:.1f}GB/{total_vram/1e9:.1f}GB, '
                      f'batch_size increase to {batch}')

        batch_ = batch
        if branching_method == 'brute-force' and num_iter <= input_split_args['bf_iters']:
            batch_ = input_split_args['bf_batch_size']

        print('Batch size:', batch_)
        try:
            global_lb = batch_verification_input_split(
                domains, net, batch_,
                num_iter=num_iter, decision_thresh=rhs, shape=x.shape,
                bounding_method=bounding_method, branching_method=branching_method,
                stop_func=stop_func, split_partitions=split_partitions)
        except AssertionError:
            if catch_assertion:
                global_lb = torch.ones(net.c.shape[0], net.c.shape[1],
                                       device=net.device) * torch.inf
            else:
                raise

        # once the lower bound stop improving we change to solve alpha mode
        if (arguments.Config["solver"]["bound_prop_method"]
            != input_split_args["enhanced_bound_prop_method"]
            and time.time() - start > input_split_args["enhanced_bound_patience"]
            and global_lb.max().cpu() <= last_glb.cpu()
            and bounding_method != "alpha-crown"
            and not enhanced_bound_initialized
        ):
            branching_method = input_split_args["enhanced_branching_method"]
            bounding_method = input_split_args["enhanced_bound_prop_method"]
            print(f'Using enhanced bound propagation method {bounding_method} '
                  f'with {branching_method} branching.')
            enhanced_bound_initialized = True

            global_lb, ret = net.build(
                init_domain, x, stop_criterion_func=stop_func(rhs),
                bounding_method=bounding_method)
            if hasattr(net.net[net.net.input_name[0]], 'lA'):
                lA = net.net[net.net.input_name[0]].lA.transpose(0, 1)
            else:
                raise ValueError("sb heuristic cannot be used without lA.")
            dm_l = x.ptb.x_L
            dm_u = x.ptb.x_U

            # compute initial split idx for the enhanced method
            split_idx = input_split_branching(
                net, global_lb, dm_l, dm_u, lA, rhs, branching_method,
                storage_depth, num_iter=num_iter)

            use_alpha = (input_split_args["enhanced_bound_prop_method"]
                            == "alpha-crown")
            domains = UnsortedInputDomainList(storage_depth, use_alpha=use_alpha)
            # This is the first batch of initial domain(s) after the branching method changed.
            domains.add(
                global_lb, dm_l.detach(), dm_u.detach(), ret['alphas'],
                net.c, rhs, split_idx)
            global_lb = global_lb.max()

        if arguments.Config["attack"]["pgd_order"] != "skip":
            if time.time() - start > input_split_args["attack_patience"]:
                print("Perform PGD attack with massively random starts finally.")
                ret_adv = massive_pgd_attack(x, net.model_ori, vnnlib=vnnlib)[1]
                if ret_adv:
                    result = "unsafe"
                    break

        if time.time() - start > timeout:
            print("Time out!")
            break

        print(f'Cumulative time: {time.time() - start}\n')
        num_iter += 1

    if result == "unknown" and len(domains) == 0:
        result = "safe"

    if return_domains:
        # Thresholds may have been updated by PGD attack so that different
        # domains may have different thresholds. Restore thresholds to the
        # default RHS for the sorting.
        domains.threshold._storage.data[:] = rhs
        domains.sort()
        if return_domains == -1:
            return_domains = len(domains)
        lower_bound, x_L, x_U = domains.pick_out_batch(
            return_domains, device='cpu')[1:4]
        return lower_bound, x_L, x_U
    else:
        del domains
        return global_lb.max(), Visited, result


def initial_verify_criterion(lbs, rhs):
    """check whether verify successful"""
    # lbs: b, n_bounds (already multiplied with c in compute_bounds())
    verified_idx = torch.any(
        (lbs - rhs) > 0, dim=-1
    )  # return bolling results in x's batch-wise
    if verified_idx.all():  # check whether all x verified
        print("Verified by initial bound!")
        return True, torch.where(verified_idx == 0)[0]
    else:
        return False, torch.where(verified_idx == 0)[0]


def transpose_c_back(lA, global_lb, rhs, dm_l, dm_u, ret, net):
    # Here, we transpose c matrix shape back from [1, spec_dim, ...] to [spec_dim, 1, ...],
    # so we should recover lA, lb, x_LB, x_UB, alphas as if they are computed with c shape [spec_dim, 1, ...],
    # to prepare for input domain bab.
    # More info can be found in function beginning comment.
    lA = lA.transpose(0, 1)
    global_lb = global_lb.transpose(0, 1)
    rhs = rhs.transpose(0, 1)
    net.c = net.c.transpose(0, 1)
    dm_l = dm_l.expand([net.c.shape[0]] + list(dm_l.shape[1:]))
    dm_u = dm_u.expand([net.c.shape[0]] + list(dm_u.shape[1:]))
    for start_node in ret['alphas']:
        for end_node in ret['alphas'][start_node]:
            if end_node == net.final_name:
                ret['alphas'][start_node][end_node] = ret[
                    'alphas'][start_node][end_node].transpose(1, 2)
            else:
                new_shape = list([
                    1 for _ in ret['alphas'][start_node][end_node].shape])
                assert ret['alphas'][start_node][end_node].shape[2] == 1
                new_shape[2] = net.c.shape[0]
                ret['alphas'][start_node][end_node] = ret[
                    'alphas'][start_node][end_node].repeat(new_shape)
    return lA, global_lb, rhs, dm_l, dm_u



from CTRAIN.train.certified.certified_training.crown_ibp import crown_ibp_train_model
from CTRAIN.train.certified.certified_training.ibp import ibp_train_model
from CTRAIN.train.certified.certified_training.shi_ibp import shi_train_model
from CTRAIN.train.certified.certified_training.sabr import sabr_train_model
from CTRAIN.train.certified.certified_training.taps import taps_train_model
from CTRAIN.train.certified.certified_training.mtl_ibp import mtl_ibp_train_model
from CTRAIN.train.certified.certified_training.staps import staps_train_model


#########################################################################
##   This file is part of the α,β-CROWN (alpha-beta-CROWN) verifier    ##
##                                                                     ##
##   Copyright (C) 2021-2024 The α,β-CROWN Team                        ##
##   Primary contacts: Huan Zhang <huan@huan-zhang.com>                ##
##                     Zhouxing Shi <zshi@cs.ucla.edu>                 ##
##                     Kaidi Xu <kx46@drexel.edu>                      ##
##                                                                     ##
##    See CONTRIBUTORS for all author contacts and affiliations.       ##
##                                                                     ##
##     This program is licensed under the BSD 3-Clause License,        ##
##        contained in the LICENCE file in this directory.             ##
##                                                                     ##
#########################################################################
import torch
from collections import defaultdict
import time
import operator
from types import SimpleNamespace
import arguments

from tensor_storage import TensorStorage
from utils import fast_hist_copy, check_infeasible_bounds


class AbstractDomainList():
    """Abstract class that maintains the list of domains (variables on CPUs)."""

    def __init__(self):
        pass

    def pick_out(self, batch, device):
        raise NotImplementedError

    def add(self, bounds, histories, depths, split,
            decision_threshs, check_infeasibility):
        raise NotImplementedError

    def get_min_domain(self, num, rev_order=False):
        # need to return a list of obj, each object has
        # lower_bounds, upper_bounds, threshold, and depth
        raise NotImplementedError

    def __len__(self):
        raise NotImplementedError

    def __getitem__(self, index):
        raise NotImplementedError

    def print(self, n=20):
        bab_args = arguments.Config["bab"]
        get_upper_bound = bab_args["get_upper_bound"]
        if len(self) > 0:
            if get_upper_bound:
                print('Current worst splitting domains [lb, ub] (depth):')
            else:
                print('Current worst splitting domains lb-rhs (depth):')
            if (arguments.Config["bab"]["cut"]["enabled"]
                    and bab_args["cut"]["cplex_cuts"]
                    and bab_args["cut"]["cplex_cuts_revpickup"]):
                printed_d = self.get_min_domain(n, rev_order=True)
            else:
                printed_d = self.get_min_domain(n)
            for i in printed_d:
                if get_upper_bound:
                    print(f'[{(i.lower_bound - i.threshold).max():.5f}, '
                          f'{(i.upper_bound - i.threshold).min():5f}] ({i.depth})', end=', ')
                else:
                    print(f'{(i.lower_bound - i.threshold).max():.5f} ({i.depth})', end=', ')
            print()


class BatchedDomainList(AbstractDomainList):
    """An unsorted but batched list of domain list."""

    def __init__(self, ret, lAs, global_lbs, global_ubs,
                 alphas=None, history=None, thresholds=None,
                 net=None, branching_input_and_activation=False, x=None):
        super().__init__()

        lb_alls, ub_alls = ret['lower_bounds'], ret['upper_bounds']
        beta = ret.get('betas', None)
        num = global_lbs.shape[0]
        assert len(lb_alls) == len(ub_alls)

        self.net = net

        self.all_lAs = {k: TensorStorage(v.cpu()) for k, v in lAs.items()}
        self.all_global_lbs = TensorStorage(global_lbs.cpu())
        self.all_global_ubs = TensorStorage(global_ubs.cpu())
        self.all_lb_alls = {k: TensorStorage(v.cpu()) for k, v in lb_alls.items()}
        self.all_ub_alls = {k: TensorStorage(v.cpu()) for k, v in ub_alls.items()}

        self.all_alphas = defaultdict(dict)
        for k in alphas:
            self.all_alphas[k] = {}
            for kk, v in alphas[k].items():
                if kk not in self.all_alphas[k]:
                    self.all_alphas[k][kk] = TensorStorage(
                        v[:, :].cpu(), concat_dim=2)
                else:
                    self.all_alphas[k][kk].append(v[:, :].cpu())
        if thresholds.numel() > 1:
            self.all_thresholds = TensorStorage(thresholds.cpu())
        else:
            thresholds = thresholds.view(1, 1)
            self.all_thresholds = TensorStorage(torch.cat([thresholds] * num).cpu())
        self.Cs = TensorStorage(net.c.cpu())

        # === separator, things above are big tensors, things below are lists ===
        self.all_betas = [beta for _ in range(num)]
        self.all_intermediate_betas = [None for _ in range(num)]
        self.histories = [fast_hist_copy(history) for _ in range(num)]
        self.split_histories = [[] for _ in range(num)]
        self.depths = [0] * num

        # tracker of number of domains
        self.l = 0
        self.u = len(self.histories)

        # === save things for statical intermediate bound ===
        self.interm_transfer = arguments.Config['bab']['interm_transfer']
        self.final_name = net.final_name
        if not self.interm_transfer:
            self.static_lb = {k: (lb[0:1].to(device='cuda', non_blocking=True)
                              if torch.cuda.is_available() else lb[0:1])
                              for k, lb in self.all_lb_alls.items()
                              if k != self.final_name}
            self.static_ub = {k: (ub[0:1].to(device='cuda', non_blocking=True)
                              if torch.cuda.is_available() else ub[0:1])
                              for k, ub in self.all_ub_alls.items()
                              if k != self.final_name}
        else:
            self.static_lb = self.static_ub = None

        self.branching_input_and_activation = branching_input_and_activation
        if branching_input_and_activation:
            input_split_idx = ret['input_split_idx']
            self.all_input_split_idx = TensorStorage(input_split_idx.cpu())
            self.all_x_Ls = TensorStorage(x.ptb.x_L.cpu())
            self.all_x_Us = TensorStorage(x.ptb.x_U.cpu())
        else:
            self.all_input_split_idx = self.all_x_Ls = self.all_x_Us = None

    def sort(self):
        """
        Sort all domains based on the margin between lower bounds and thresholds.
        """
        sort_time = time.time()
        N = self.u
        if N <= 0:
            return
        global_lb = self.all_global_lbs[:N]
        thresholds = self.all_thresholds[:N]
        indices = (global_lb - thresholds).max(dim=1).values.argsort(descending=True)
        indices_lst = indices.tolist()

        # Reorder all single tensor storage items or list of items.
        for item in [
            self.all_x_Ls, self.all_x_Us, self.all_input_split_idx,
            self.all_global_lbs, self.all_global_ubs, self.Cs,
            self.all_thresholds
        ] + (list(self.all_lAs.values()) + list(self.all_lb_alls.values())
             + list(self.all_ub_alls.values())):
            if item is not None:
                item._storage[:N] = item._storage[indices]

        # Reorder alphas (dictionary of dictionary).
        for v in self.all_alphas.values():
            for vv in v.values():
                # The batch dimension is at 2 for alphas.
                vv._storage[:,:,:N,:] = (
                    vv._storage[:,:,:N,:].index_select(index=indices, dim=2))

        # Reorder lists.
        selector = operator.itemgetter(*indices_lst) if len(
                indices_lst) > 1 else lambda _arr: (_arr[indices_lst[0]], )
        for name in ('histories', 'depths', 'split_histories',
                'all_betas', 'all_intermediate_betas'):
            item = getattr(self, name)
            if item is not None:
                setattr(self, name, list(selector(item)))
        sort_time = time.time() - sort_time
        print(f'Sorting batched domains takes {sort_time} seconds.')

    def pick_out(self, batch, device='cpu', impl_params=None):
        def _to(x, non_blocking=True):
            return x.to(device=device, non_blocking=non_blocking)

        if torch.cuda.is_available():
            torch.cuda.synchronize()  # make sure GPU to CPU transfer is finished

        assert batch > 0
        batch = min(len(self), int(batch))

        new_lAs = {k: _to(lA.pop(batch)) for k, lA in self.all_lAs.items()}
        new_x_Ls = _to(self.all_x_Ls.pop(batch)) if self.all_x_Ls is not None else None
        new_x_Us = _to(self.all_x_Us.pop(batch)) if self.all_x_Us is not None else None
        new_input_split_idx = (_to(self.all_input_split_idx.pop(batch))
                               if self.all_input_split_idx is not None else None)
        global_lb = _to(self.all_global_lbs.pop(batch))
        alphas = defaultdict(dict)
        for k, v_map in self.all_alphas.items():
            alphas[k] = {kk: _to(vv.pop(batch)) for kk, vv in v_map.items()}
        betas_all = self.all_betas[self.u - batch: self.u]
        intermediate_betas_all = self.all_intermediate_betas[self.u - batch: self.u]
        cs = _to(self.Cs.pop(batch))
        thresholds = _to(self.all_thresholds.pop(batch))

        if self.interm_transfer:
            lower_bounds = {k: _to(v.pop(batch))
                            for k, v in self.all_lb_alls.items()}
            upper_bounds = {k: _to(v.pop(batch))
                            for k, v in self.all_ub_alls.items()}
        else:
            lower_bounds, upper_bounds = self._interm_bounds_wo_transfer(
                batch, device, impl_params)
        # TODO Move to beta_CROWN_solver. Duplicate code.
        new_masks = {}
        for k in lower_bounds:
            if k not in self.net.split_activations:
                continue
            mask = None
            for activation, index in self.net.split_activations[k]:
                mask_ = _to(
                    activation.get_split_mask(
                        lower_bounds[k], upper_bounds[k], index
                    ).flatten(1).float())
                mask = mask_ if mask is None else torch.logical_or(mask, mask_)
            if mask is None:
                mask = torch.ones_like(lower_bounds[k]).flatten(1)
            new_masks[k] = mask

        history = self.histories[self.u - batch: self.u]
        split_history = self.split_histories[self.u - batch: self.u]
        depths = self.depths[self.u - batch: self.u]
        self.u -= batch

        self.all_betas = self.all_betas[:self.u]
        self.all_intermediate_betas = self.all_intermediate_betas[:self.u]
        self.histories = self.histories[:self.u]
        self.split_histories = self.split_histories[:self.u]
        self.depths = self.depths[:self.u]

        return {
            'mask': new_masks, 'lAs': new_lAs,
            'lower_bounds': lower_bounds, 'upper_bounds': upper_bounds,
            'alphas': alphas, 'betas': betas_all,
            'intermediate_betas': intermediate_betas_all,
            'history': history, 'split_history': split_history,
            'global_lb': global_lb, 'depths': depths, 'cs': cs,
            'thresholds': thresholds, 'x_Ls': new_x_Ls, 'x_Us': new_x_Us,
            'input_split_idx': new_input_split_idx,
        }

    def add(self, bounds, d, check_infeasibility):
        histories = d['history']
        decision_threshs = d['thresholds']
        final_lower = bounds['lower_bounds'][self.final_name]
        batch = final_lower.size(0)
        device = final_lower.device
        decision_threshs = decision_threshs.to(device, non_blocking=True)
        assert (self.all_x_Ls is None) == (bounds['x_Ls'] is None), "Inconsistent x_Ls during construction and using {type(self)}."
        assert (self.all_x_Us is None) == (bounds['x_Us'] is None), "Inconsistent x_Ls during construction and using {type(self)}."
        assert (self.all_input_split_idx is None) == (bounds['input_split_idx'] is None), "Inconsistent input_split_idx during construction and using {type(self)}"
        assert len(self.all_lAs) == len(bounds['lAs'])

        if check_infeasibility:
            infeasible = check_infeasible_bounds(
                bounds['lower_bounds'], bounds['upper_bounds'])
        else:
            infeasible = None

        # torch.all() is when taking max for multiple output specifications
        indexer = torch.all(
            bounds['lower_bounds'][self.final_name] <= decision_threshs[:batch],
            dim=1)
        if infeasible is not None:
            indexer = torch.logical_and(indexer, torch.logical_not(infeasible))
        indexer = indexer.nonzero().view(-1)
        if len(indexer) == 0:
            return
        # Add all list items in batch without for loop.
        if len(indexer) > 0:
            # itemgetter returns a value instead of tuple when length is 1, so need a special case.
            batch_indexer_lst = indexer.tolist()
            selector = (
                operator.itemgetter(*batch_indexer_lst)
                if len(batch_indexer_lst) > 1
                else lambda _arr: (_arr[batch_indexer_lst[0]], ))
            self.histories.extend(selector(histories))
            self.all_betas.extend(selector(bounds['betas']))
            self.all_intermediate_betas.extend(selector(bounds['intermediate_betas']))
            self.split_histories.extend(selector(bounds['split_history']))
            self.depths.extend(selector(d['depths']))
        for k in self.all_lAs:
            self.all_lAs[k].append(bounds['lAs'][k][indexer])
        if self.all_x_Ls is not None:
            self.all_x_Ls.append(bounds['x_Ls'][indexer])
        if self.all_x_Us is not None:
            self.all_x_Us.append(bounds['x_Us'][indexer])
        if self.all_input_split_idx is not None:
            self.all_input_split_idx.append(bounds['input_split_idx'][indexer])
        self.all_global_lbs.append(bounds['lower_bounds'][self.final_name][indexer])
        self.all_global_ubs.append(bounds['upper_bounds'][self.final_name][indexer])
        for k, v in bounds['lower_bounds'].items():
            self.all_lb_alls[k].append(v[indexer])
        for k, v in bounds['upper_bounds'].items():
            self.all_ub_alls[k].append(v[indexer])
        alpha_new = alpha_reuse = False
        for k, v in bounds['alphas'].items():
            if k not in self.all_alphas:
                self.all_alphas[k] = {}
            for kk, vv in v.items():
                if kk not in self.all_alphas[k]:
                    # This is the first time to create these alpha TensorStorage
                    self.all_alphas[k][kk] = TensorStorage(vv[:,:,indexer].cpu(), concat_dim=2)
                    alpha_new = True
                else:
                    # Reusing existing TensorStorage
                    self.all_alphas[k][kk].append(vv[:,:,indexer])
                    alpha_reuse = True
        assert not (alpha_new and alpha_reuse)
        self.all_thresholds.append(decision_threshs[indexer])
        self.Cs.append(bounds['c'][indexer])
        self.u = len(self.histories)

    def _interm_bounds_wo_transfer(self, batch, device, impl_params):
        def _to(x, non_blocking=True):
            return x.to(device=device, non_blocking=non_blocking)

        # place dummy place to record intermediate bounds
        lower_bounds, upper_bounds = {}, {}
        lower_bounds[self.final_name] = _to(
            self.all_lb_alls[self.final_name].pop(batch))
        upper_bounds[self.final_name] = _to(
            self.all_ub_alls[self.final_name].pop(batch))
        # FIXME This part of code looks unreliable. Need to be clear when
        # this situation might happen.
        # repeat static_lb and static_ub when necessary
        for k in self.static_lb:
            # enlarge the batch size in the static storage
            if batch > self.static_lb[k].shape[0]:
                power = (batch + self.static_lb[k].shape[0] - 1) // self.static_lb[k].shape[0]
                self.static_lb[k] = self.static_lb[k].repeat(
                    power, *tuple([1] * (self.static_lb[k].dim() - 1)))
            if batch > self.static_ub[k].shape[0]:
                power = (batch + self.static_ub[k].shape[0] - 1) // self.static_ub[k].shape[0]
                self.static_ub[k] = self.static_ub[k].repeat(
                    power, *tuple([1] * (self.static_ub[k].dim() - 1)))
        # need to fill in the slots
        histories = self.histories[self.u - batch: self.u]
        # setting positive and negative neurons with lb 0 or ub 0
        lb_row_selector = {k: [] for k in self.static_lb}
        lb_col_selector = {k: [] for k in self.static_lb}
        ub_row_selector = {k: [] for k in self.static_lb}
        ub_col_selector = {k: [] for k in self.static_lb}

        for k in self.static_lb:
            lower_bounds[k] = self.static_lb[k][:batch].clone()
            upper_bounds[k] = self.static_ub[k][:batch].clone()

        if impl_params is not None:
            pass
        else:
            # TODO merge the two loops
            for i in self.static_lb:
                for j, hist in enumerate(histories):
                    if isinstance(hist[i][0], torch.Tensor):
                        hist0 = hist[i][0].__array__()
                    else:
                        hist0 = hist[i][0]
                    if isinstance(hist[i][1], torch.Tensor):
                        hist1 = hist[i][1].__array__()
                    else:
                        hist1 = hist[i][1]
                    for k in range(len(hist0)):
                        idx = hist0.item(k)
                        direc = hist1.item(k)
                        if direc >= 0:
                            lb_row_selector[i].append(j)
                            lb_col_selector[i].append(idx)
                        else:
                            ub_row_selector[i].append(j)
                            ub_col_selector[i].append(idx)
                if len(lb_row_selector[i]) > 0:
                    lower_bounds[i].view(batch, -1)[
                        lb_row_selector[i], lb_col_selector[i]] = 0.0
                if len(ub_row_selector[i]) > 0:
                    upper_bounds[i].view(batch, -1)[
                        ub_row_selector[i], ub_col_selector[i]] = 0.0

        return lower_bounds, upper_bounds

    def _assemble_domains(self, global_lbs, global_ubs, history, split_history, depth, thresholds):
        ans = []
        for lb, ub, now_his, now_split_his, now_depth, now_threshold in zip(global_lbs, global_ubs, history, split_history, depth, thresholds):
            now_obj = SimpleNamespace()
            now_obj.history = now_his
            now_obj.split_history = now_split_his
            now_obj.depth = now_depth
            now_obj.lower_bound = lb
            now_obj.upper_bound = ub
            now_obj.threshold = now_threshold
            ans.append(now_obj)
        return ans

    def get_min_domain(self, num, rev_order=False):
        indices = (self.all_global_lbs - self.all_thresholds).max(dim=1)[0].argsort()[:num]
        indices_lst = indices.tolist()

        selected_candidate_domains = self._assemble_domains(
            self.all_global_lbs[indices],
            self.all_global_ubs[indices],
            [self.histories[i] for i in indices_lst],
            [self.split_histories[i] for i in indices_lst],
            [self.depths[i] for i in indices_lst],
            self.all_thresholds[indices])
        return selected_candidate_domains

    def __len__(self):
        return self.u - self.l

    def __getitem__(self, index):
        selected_candidate_domains = self._assemble_domains(
            self.all_global_lbs[index: index + 1],
            self.all_global_ubs[index: index + 1],
            [self.histories[index]],
            [self.split_histories[index]],
            [self.depths[index]],
            self.all_thresholds[index: index + 1])
        return selected_candidate_domains[0]


def check_worst_domain(d):
    use_cuts = (arguments.Config["bab"]["cut"]["enabled"]
                and arguments.Config["bab"]["cut"]["cplex_cuts"]
                and arguments.Config["bab"]["cut"]["cplex_cuts_revpickup"])
    if len(d) > 0:
        if use_cuts:
            worst_domain = d.get_min_domain(1, rev_order=True)
            global_lb = worst_domain[-1].lower_bound - worst_domain[-1].threshold
        else:
            worst_domain = d.get_min_domain(1, rev_order=False)
            global_lb = worst_domain[0].lower_bound - worst_domain[0].threshold
        return global_lb
    else:
        return torch.tensor(1e-7)



#########################################################################
##   This file is part of the α,β-CROWN (alpha-beta-CROWN) verifier    ##
##                                                                     ##
##   Copyright (C) 2021-2024 The α,β-CROWN Team                        ##
##   Primary contacts: Huan Zhang <huan@huan-zhang.com>                ##
##                     Zhouxing Shi <zshi@cs.ucla.edu>                 ##
##                     Kaidi Xu <kx46@drexel.edu>                      ##
##                                                                     ##
##    See CONTRIBUTORS for all author contacts and affiliations.       ##
##                                                                     ##
##     This program is licensed under the BSD 3-Clause License,        ##
##        contained in the LICENCE file in this directory.             ##
##                                                                     ##
#########################################################################
r"""
This file is an example of defining a custom operation and providing its
relaxations for bound computation. Here we consider a modified activation
function sigma which is a mixture between a linear function and non-linear
function:
             / a_i x_i + b_i   if m_i = 1,
    f(x_i) = |
             \ sigmoid(x_i)       if m_i = 0.
where a, b are element-wise slopes and biases when the function is linear,
and m is the mask controlling the behavior of this function. We consider
perturbations on x.

This example is a slight extension of `custom_op.py`.
"""
import torch
import torch.nn as nn
from auto_LiRPA import register_custom_op
from auto_LiRPA.bound_ops import BoundSigmoid, Interval


class LinearMaskedSigmoidOp(torch.autograd.Function):
    """A sigmoid function with some neurons replaced with linear operations."""
    @staticmethod
    def forward(ctx, input: torch.Tensor, mask: torch.Tensor, slope: torch.Tensor, bias: torch.Tensor) -> torch.Tensor:
        # mask = 1 => using linear operation input * slope + bias, mask = 0 => using Sigmoid
        s = torch.sigmoid(input)
        ctx.save_for_backward(input, mask, slope, bias, s)
        return s * (1.0 - mask) + (input * slope + bias) * mask

    def backward(ctx, grad_output):
        """
        In the backward pass we receive a Tensor containing the gradient of the loss
        with respect to the output, and we need to compute the gradient of the loss
        with respect to the input.
        """
        input, mask, slope, bias, s = ctx.saved_tensors
        sigmoid_grad = grad_output * s * (1.0 - s)
        grad_input = sigmoid_grad * (1.0 - mask) + grad_output * mask * slope
        grad_slope = grad_output * input * mask
        grad_bias = grad_output * mask
        grad_mask = -s + (input * slope + bias)
        return grad_input, grad_mask, grad_slope, grad_bias

    @staticmethod
    def symbolic(g, input, mask, weight, bias):
        # This will be parsed as a custom operation when doing the ONNX conversion.
        return g.op("customOp::LinearMaskedSigmoid", input, mask, weight, bias)


class LinearMaskedSigmoid(nn.Module):
    """Create a module to wrap the parameters for LinearMaskedSigmoidOp."""
    def __init__(self, size):
        super().__init__()
        if isinstance(size, int):
            size = (size, )
        # All mask, slope and bias are element-wise.
        self.register_buffer('mask', (torch.rand(size=size) > 0).to(dtype=torch.get_default_dtype()))
        self.register_buffer('alpha', torch.rand(size=size, dtype=torch.get_default_dtype()))
        self.register_buffer('bias', torch.rand(size=size, dtype=torch.get_default_dtype()))

    def forward(self, input):
        # mask = 1 => using linear operation input * slope + bias, mask = 0 => using Sigmoid.
        return LinearMaskedSigmoidOp.apply(input, self.mask, self.slope, self.bias)


class BoundLinearMaskedSigmoid(BoundSigmoid):
    """This class defines how we compute the bounds for our customized Sigmoid function."""

    def forward(self, x, mask=None, slope=None, bias=None):
        """Regular forward propagation (e.g., for evaluating clean accuracy)."""
        if mask is None or slope is None or bias is None:
            # If mask is not given, then just act like a sigmoid.
            return super().forward(x)
        return LinearMaskedSigmoidOp.apply(x, mask, slope, bias)

    def interval_propagate(self, x, mask, slope, bias):
        """Interval bound propagation (IBP)."""
        # Each x, mask, slope, bias is a tuple, or a Interval object representing lower and upper bounds.
        # We assume Linf norm perturbation on input.
        assert Interval.get_perturbation(x)[0] == float("inf")
        x_L, x_U = x[0], x[1]  # The inputs (x)
        # We assume no perturbations on mask, slope and bias.
        mask, slope, bias = mask[0], slope[0], bias[0]
        # Lower and upper bounds when Sigmoid is selected.
        sigmoid_lb = torch.sigmoid(x_L)
        sigmoid_ub = torch.sigmoid(x_U)
        # Lower and upper bounds when linear coefficients are selected.
        pos_slope = (slope >= 0).to(dtype=torch.get_default_dtype())
        neg_slope = 1.0 - pos_slope
        linear_lb = pos_slope * slope * x_L + neg_slope * slope * x_U + bias
        linear_ub = pos_slope * slope * x_U + neg_slope * slope * x_L + bias
        # Select the final bounds according to the mask.
        final_lb = mask * linear_lb + (1.0 - mask) * sigmoid_lb
        final_ub = mask * linear_ub + (1.0 - mask) * sigmoid_ub
        return final_lb, final_ub

    def bound_relax(self, x, init=False, dim_opt=None):
        """Element-wise CROWN relaxation for our special Sigmoid activation function."""
        # Call parent class to relax Sigmoid neurons.
        super().bound_relax(x, init=init, dim_opt=dim_opt)
        # Modify the relaxation coefficients for these linear neurons.
        neg_mask = 1.0 - self._mask
        masked_slope = self._mask * self._slope
        masked_bias = self._mask * self._bias
        self.uw = masked_slope + neg_mask * self.uw
        self.ub = masked_bias + neg_mask * self.ub
        self.lw = masked_slope + neg_mask * self.lw
        self.lb = masked_bias + neg_mask * self.lb

    def bound_backward(self, last_lA, last_uA, x, mask, slope, bias, **kwargs):
        """Backward LiRPA (CROWN) bound propagation."""
        # These are additional variables that will be used in _backward_relaxation(), so we save them here.
        self._mask = mask.buffer  # These are registered as buffers; see class BoundBuffer.
        self._slope = slope.buffer
        self._bias = bias.buffer
        # The parent class will call _backward_relaxation() and obtain the relaxations,
        # and that's all we need; after obtaining linear relaxations for each neuron, other
        # parts of class BoundSigmoid can be reused.
        As, lbias, ubias = super().bound_backward(last_lA, last_uA, x, **kwargs)
        # Returned As = [(lA, uA)]; these A matrices are for input x.
        # Our customized Sigmoid has three additional buffers as inputs; we need to set their
        # corresponding A matrices to None. The length of As must match the number of inputs
        # of this customize function.
        As += [(None, None), (None, None), (None, None)]
        return As, lbias, ubias


class linear_masked_sigmoid_model(nn.Module):
    """Model for testing."""
    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv2d(3, 16, 4, stride=2, padding=0)
        # Using our customized Sigmoid function to replace the original Sigmoid function.
        self.linear_masked_sigmoid1 = LinearMaskedSigmoid(size=(16,15,15))
        self.conv2 = nn.Conv2d(16, 32, 4, stride=2, padding=0)
        self.linear_masked_sigmoid2 = LinearMaskedSigmoid(size=(32,6,6))
        self.flatten = nn.Flatten()
        self.linear1 = nn.Linear(32*6*6,100)
        self.linear_masked_sigmoid3 = LinearMaskedSigmoid(100)
        self.linear2 = nn.Linear(100, 10)
        # Register the customized op in auto_LiRPA library.
        register_custom_op("customOp::LinearMaskedSigmoid", BoundLinearMaskedSigmoid)

    def forward(self, x):
        out = self.conv1(x)
        out = self.linear_masked_sigmoid1(out)
        # out = torch.nn.functional.sigmoid(out)
        out = self.conv2(out)
        out = self.linear_masked_sigmoid2(out)
        # out = torch.nn.functional.sigmoid(out)
        out = self.flatten(out)  # Flatten must be after activation for most efficient computation.
        out = self.linear1(out)
        out = self.linear_masked_sigmoid3(out)
        # out = torch.nn.functional.sigmoid(out)
        out = self.linear2(out)
        return out




#########################################################################
##   This file is part of the α,β-CROWN (alpha-beta-CROWN) verifier    ##
##                                                                     ##
##   Copyright (C) 2021-2024 The α,β-CROWN Team                        ##
##   Primary contacts: Huan Zhang <huan@huan-zhang.com>                ##
##                     Zhouxing Shi <zshi@cs.ucla.edu>                 ##
##                     Kaidi Xu <kx46@drexel.edu>                      ##
##                                                                     ##
##    See CONTRIBUTORS for all author contacts and affiliations.       ##
##                                                                     ##
##     This program is licensed under the BSD 3-Clause License,        ##
##        contained in the LICENCE file in this directory.             ##
##                                                                     ##
#########################################################################
import arguments
import os
import csv
import re
import torch
import numpy as np
from load_model import load_model, load_model_onnx, Customized  # pylint: disable=unused-import
from read_vnnlib import read_vnnlib
from data_utils import load_eran_dataset, load_sdp_dataset
from data_utils import load_sampled_dataset, load_generic_dataset
from data_utils import load_pkl_dataset
from utils import get_save_path, expand_path
from specifications import construct_vnnlib


def default_onnx_and_vnnlib_loader(file_root, onnx_path, vnnlib_path):
    vnnlib = read_vnnlib(os.path.join(file_root, vnnlib_path))
    model_ori, onnx_shape = load_model_onnx(
        os.path.join(file_root, onnx_path), x=torch.tensor(vnnlib[0][0])[:, 0])
    shape = (-1, *onnx_shape)  # add the batch dim to onnx_shape
    model_ori, vnnlib = joint_optimization_with_onnx_and_vnnlib(model_ori, shape, vnnlib,
                                                                arguments.Config['model']['onnx_vnnlib_joint_optimization_flags'])
    return model_ori, shape, vnnlib


# TODO (2/17/23): Move it `loading.py` after the cleanup.
def load_verification_dataset():
    spec = arguments.Config['specification']

    if arguments.Config["data"]["pkl_path"] is not None:
        dataset_method = load_pkl_dataset
    elif arguments.Config['data']['dataset'].startswith('Customized('):
        dataset_method = eval(arguments.Config["data"]["dataset"])  # pylint: disable=eval-used
    elif "ERAN" in arguments.Config["data"]["dataset"] or "MADRY" in arguments.Config["data"]["dataset"]:
        dataset_method = load_eran_dataset
    elif "SDP" in arguments.Config["data"]["dataset"]:
        dataset_method = load_sdp_dataset
    elif "SAMPLE" in arguments.Config["data"]["dataset"]:
        # Sampled datapoints (a small subset of MNIST/CIFAR), only for reproducing some paper results.
        dataset_method = load_sampled_dataset
    elif "CIFAR" in arguments.Config["data"]["dataset"] or "MNIST" in arguments.Config["data"]["dataset"]:
        dataset_method = load_generic_dataset
    else:
        raise NotImplementedError(
            'Dataset not supported in this file! '
            'Please customize load_verification_dataset() function in utils.py.')

    # FIXME (01/10/22): fully document customized data loader.
    # Returns: X, labels, runnerup, data_max, data_min, eps, target_label.
    # X is the data matrix in (batch, ...).
    # labels are the ground truth labels, a tensor of integers.
    # runnerup is the runnerup label used for quickly verify against the runnerup (second largest) label, can be set to None.
    # data_max is the per-example perturbation upper bound, shape (batch, ...) or (1, ...).
    # data_min is the per-example perturbation lower bound, shape (batch, ...) or (1, ...).
    # eps is the Lp norm perturbation epsilon. Can be set to None if element-wise perturbation (specified by data_max and data_min) is used.
    # Target label is the targeted attack label; can be set to None.
    data_config = dataset_method(spec)
    if isinstance(data_config, dict):
        ret = data_config
    else:
        # TODO (2/17/2023) deprecate
        runnerup, target_labels = None, None
        if len(data_config) == 5:
            X, labels, data_max, data_min, eps_new = data_config
        elif len(data_config) == 6:
            X, labels, data_max, data_min, eps_new, runnerup = data_config
        elif len(data_config) == 7:
            X, labels, data_max, data_min, eps_new, runnerup, target_labels = data_config
        else:
            raise NotImplementedError("Data config types not correct!")
        ret = {
            'X': X, 'labels': labels, 'runnerup': runnerup,
            'data_max': data_max, 'data_min': data_min,
            'eps': eps_new, 'norm': arguments.Config['specification']['norm'],
            'target_label': target_labels
        }
    if 'labels' in ret:
        assert ret['X'].size(0) == ret['labels'].size(0), "batch size of X and labels should be the same!"
    if 'data_max' in ret and 'data_min' in ret:
        assert (ret['data_max'] - ret['data_min']).min() >= 0, "data_max should always be larger or equal to data_min!"
    if arguments.Config["specification"]["norm"] != np.inf:
        if isinstance(arguments.Config["data"]["std"], (list, tuple)):
            assert arguments.Config["data"]["std"].count(arguments.Config["data"]["std"][0]) == len(
                arguments.Config["data"]["std"]), ('For non-Linf norm, we support only 1-d eps (all channels with the same perturbation). '
                'If you have more complex, per-channel eps (e.g., an ellipsoid L2 perturbation, you can '
                'add the data normalization into part of the model.')
            arguments.Config["data"]["std"] = arguments.Config["data"]["std"][0]
        else:
             arguments.Config["data"]["std"] = float(arguments.Config["data"]["std"])
        # FIXME need to ensure all the values are equal in the tensor
        if isinstance(ret['eps'], torch.Tensor):
            ret['eps'] = ret['eps'][0, 0, 0, 0]  # only support eps as a scalar for non-Linf norm

    return ret


def load_model_and_vnnlib(file_root, csv_item):
    onnx_path, vnnlib_path, arguments.Config["bab"]["timeout"] = csv_item
    onnx_path = os.path.join(
        arguments.Config["model"]["onnx_path_prefix"], onnx_path.strip())
    vnnlib_path = os.path.join(
        arguments.Config["specification"]["vnnlib_path_prefix"],
        vnnlib_path.strip())
    print(f'Using onnx {onnx_path}')
    print(f'Using vnnlib {vnnlib_path}')
    model_ori, shape, vnnlib = eval(  # pylint: disable=eval-used
        arguments.Config["model"]["onnx_loader"]
    )(file_root, onnx_path, vnnlib_path)
    return model_ori, shape, vnnlib, onnx_path


def adhoc_tuning(data_min, data_max, model_ori):
    if 'vgg' in arguments.Config['general']['root_path']:
        perturbed = (data_max - data_min > 0).sum()
        if perturbed > 10000:
            print('WARNING: prioritizing attack due to too many perturbed pixels on VGG')
            print('Setting arguments.Config["attack"]["pgd_order"] to "before"')
            arguments.Config['attack']['pgd_order'] = 'before'

    if 'nn4sys' in arguments.Config['general']['root_path']:
        if data_max.shape == torch.Size([1, 1]):
            print(f'Enlarging initial_max_domains for model with input shape {data_max.shape}')
            arguments.Config['bab']['initial_max_domains'] = 100000

    if 'vit' in arguments.Config['general']['root_path']:
        num_matmul = len([item for item in model_ori.modules()
                      if 'MatMul' in str(type(item))])
        if num_matmul >= 6:
            print('Sharing alpha due to model size')
            arguments.Config['solver']['alpha-crown']['share_alphas'] = True


def parse_run_mode():
    """ parse running by vnnlib or customized data
     if using customized data, we convert them to vnnlib format
     """
    model_ori = vnnlib_all = shape = None
    file_root = expand_path(arguments.Config['general']['root_path'])
    if (arguments.Config['general']['csv_name']
            and arguments.Config['specification']['vnnlib_path'] is None):
        # A CSV filename is specified, and we will go over all models and specs in this csv file.
        # Used for running VNN-COMP benchmarks in batch mode.
        # In this case, vnnlib_path should not be specified, otherwise we will run only a single model/spec.
        run_mode = 'csv_file'
        csv_path = os.path.join(file_root, arguments.Config['general']['csv_name'])
        with open(csv_path, newline='') as csv_f:
            reader = csv.reader(csv_f, delimiter=',')
            # In VNN-COMP each line of the csv contains 3 elements: model, vnnlib, timeout
            csv_file = [row for row in reader]

        if len(csv_file[0]) == 1:
            # Each row contains only one item, which is the vnnlib spec. So we load and return the model only once here.
            # This case is used when we have a batch of vnnlib specs to verify with one model either pytorch or ONNX.
            model_ori = load_model()

        arguments.Config['data']['end'] = min(arguments.Config['data']['end'], reader.line_num)
        if arguments.Config['data']['start'] != 0 or arguments.Config['data']['end'] != reader.line_num:
            assert (0 <= arguments.Config['data']['start'] <= reader.line_num
                    and arguments.Config['data']['end'] > arguments.Config['data']['start']
            ), 'specified --start or --end out of range: start={}, end={}, total_in_csv={}'.format(
                arguments.Config['data']['end'],
                arguments.Config['data']['start'], reader.line_num)
            print('customized start/end sample from instance {} to {} in {}'.format(
                arguments.Config['data']['start'], arguments.Config['data']['end'],
                arguments.Config['general']['csv_name']))
        else:
            print('no customized start/end sample, testing all samples in {}'.format(
                arguments.Config['general']['csv_name']))
            arguments.Config['data']['start'], arguments.Config['data']['end'] = 0, reader.line_num
        example_idx_list = csv_file[arguments.Config['data']['start']:arguments.Config['data']['end']]
    elif arguments.Config['model']['onnx_path'] is not None and arguments.Config['specification']['vnnlib_path'] is not None:
        # A onnx file and a vnnlib file is specified, run this onnx file with vnnlib, ignore csv file.
        # Used for VNN-COMP in single instance mode, will be used in run_instance.sh
        run_mode = 'single_vnnlib'
        arguments.Config['data']['start'], arguments.Config['data']['end'] = 0, 1
        csv_file = [(arguments.Config['model']['onnx_path'],
                     arguments.Config['specification']['vnnlib_path'],
                     arguments.Config['bab']['timeout'])]
        file_root = ''
        example_idx_list = csv_file[arguments.Config['data']['start']:arguments.Config['data']['end']]
    elif not arguments.Config['general']['csv_name']:
        # No CSV specified, we will use specifications defined in yaml file.
        # This part replaces the old robustness_verifier.py interface.
        run_mode = 'customized_data'
        # Load Pytorch or ONNX model depends on the model path or onnx_path is given.
        model_ori = load_model(weights_loaded=True)
        if arguments.Config['specification']['vnnlib_path'] is None:
            verification_dataset = load_verification_dataset()
            X = verification_dataset['X']
            # X, labels, runnerup, data_max, data_min, perturb_epsilon, target_label = verification_dataset
            if arguments.Config['data']['data_idx_file'] is not None:
                # Go over a list of data indices.
                with open(arguments.Config['data']['data_idx_file']) as f:
                    example_idx_list = re.split(r'[;|,|\n|\s]+', f.read().strip())
                    example_idx_list = [int(b_id) for b_id in example_idx_list]
                    print(f'Example indices (total {len(example_idx_list)}): {example_idx_list}')
            else:
                # By default, we go over all data.
                example_idx_list = list(range(X.shape[0]))
            example_idx_list = example_idx_list[arguments.Config['data']['start']:arguments.Config['data']['end']]
            vnnlib_all = construct_vnnlib(verification_dataset, example_idx_list)
            # X, labels, runnerup, data_max, data_min, perturb_epsilon,
            # target_label, example_idx_list)
            shape = [-1] + list(X.shape[1:])
        else:
            # Using vnnlib specification (e.g., loading a pytorch model and use vnnlib to define general specification).
            example_idx_list = [0]
            vnnlib = read_vnnlib(arguments.Config['specification']['vnnlib_path'])
            assert arguments.Config['model']['input_shape'] is not None, 'vnnlib does not have shape information, please specify by --input_shape'
            shape = arguments.Config['model']['input_shape']
            vnnlib_all = [vnnlib]  # Only 1 vnnlib file.
    else:
        raise NotImplementedError

    if arguments.Config['general']['results_file']:
        save_path = arguments.Config['general']['results_file']
    else:
        save_path = get_save_path(
            csv=arguments.Config['general']['csv_name'] is not None)

    print(f'Internal results will be saved to {save_path}.')
    # FIXME_NOW: model_ori should not be handled in this function! Do it in the utility function that loads models for all cases.
    return run_mode, save_path, file_root, example_idx_list, model_ori, vnnlib_all, shape


def joint_optimization_with_onnx_and_vnnlib(model, shape, vnnlib, flags):
    for opt_flag in flags:
        if opt_flag == "none":
            continue
        elif opt_flag == "peel_off_last_softmax_layer":
            def inv_sigmoid(c, rhs, optype):
                nonzero_c = [c[0][i] for i in c[0].nonzero()[0]]
                if optype == 'Sigmoid':
                    if len(nonzero_c) == 1:
                        # [ 0 ... 0 k 0 ...] <> rhs <=> ksigmoid(x_i) <> rhs <=> k x_i <> k * sigmoid^(-1) (rhs/k)
                        np_rhs = np.array(rhs)
                        np_c = np.array(c).reshape(np_rhs.shape)
                        # y = 1/(1+exp(-x)) => x = ln ( y / (1-y) )
                        new_rhs = np.log((np_rhs / np_c) / (- (np_rhs / np_c) + 1.)) * np_c
                        return new_rhs.tolist()
                    if len(nonzero_c) == 2 and 1 in nonzero_c and -1 in nonzero_c and rhs[0] == 0:
                        # [0 .. 0 1 0 ... 0 -1 0 ...] <> 0 <=> sigmoid(x_i) - sigmoid(x_j) <> 0 <=> x_i - x_j <> 0
                        return rhs
                elif optype == 'Softmax':
                    # [0 .. 0 1 0 ... 0 -1 0 ...] <> 0 <=> softmax(x_i) - softmax(x_j) <> 0 <=> x_i - x_j <> 0
                    assert len(nonzero_c) == 2 and 1 in nonzero_c and -1 in nonzero_c and rhs[0] == 0, \
                        f"Unsupported (c,rhs) for removing Softmax: ({c}, {rhs})"
                    return rhs
                else:
                    raise NotImplementedError

            ori_output_nodename = model.output_names[0]
            ori_output_nodeidx = 0
            new_output_nodename = None
            removed_op_type = None
            for node in model.onnx_model.graph.node:
                if (node.op_type == 'Sigmoid' or node.op_type == 'Softmax') and node.output[0] == ori_output_nodename:
                    new_output_nodename = node.input[0]
                    removed_op_type = node.op_type
                    break
                ori_output_nodeidx += 1

            if new_output_nodename is not None:
                print(f'Last {removed_op_type} node found, peel it off')
                # delete from onnx model
                model.output_names = [new_output_nodename]
                del model.onnx_model.graph.node[ori_output_nodeidx]
                # update the rhs in vnnlib
                new_vnnlib = [(vlib[0], [(c, inv_sigmoid(c, rhs, removed_op_type)) for c, rhs in vlib[1]]) for vlib in vnnlib]
                vnnlib = new_vnnlib

    return model, vnnlib



#########################################################################
##   This file is part of the α,β-CROWN (alpha-beta-CROWN) verifier    ##
##                                                                     ##
##   Copyright (C) 2021-2024 The α,β-CROWN Team                        ##
##   Primary contacts: Huan Zhang <huan@huan-zhang.com>                ##
##                     Zhouxing Shi <zshi@cs.ucla.edu>                 ##
##                     Kaidi Xu <kx46@drexel.edu>                      ##
##                                                                     ##
##    See CONTRIBUTORS for all author contacts and affiliations.       ##
##                                                                     ##
##     This program is licensed under the BSD 3-Clause License,        ##
##        contained in the LICENCE file in this directory.             ##
##                                                                     ##
#########################################################################
"""
This file shows how to use a customized model loader, and it defines a
few customized loaders for VNN-COMP models as examples. See example
configuration files in the `exp_configs` folder, e.g.,
`exp_configs/vnncomp21/marabou_cifar10.yaml`.
"""

import os
import torch
import torch.nn as nn
import numpy as np
import onnx
from onnx import numpy_helper as nh
import onnx2pytorch
import arguments

from onnx_opt import create_initializer_tensor
from load_model import load_model_onnx
from loading import joint_optimization_with_onnx_and_vnnlib
from read_vnnlib import read_vnnlib
from model_defs import Step_carvana


def transpose_nhwc(model_ori, vnnlib, shape):
    """Transpose vnnlib when NHWC is used (marabou-cifar10)."""
    # Only sequential models are supported and there must be a transpose
    # layer in the beginning.
    assert isinstance(model_ori, nn.Sequential)
    assert 'Transpose' in str(type(model_ori._modules['0']))
    model_ori = nn.Sequential(*(list(model_ori._modules.values())[1:]))
    assert len(vnnlib) == 1 and len(vnnlib[0]) == 2
    x = torch.tensor(vnnlib[0][0])
    assert x.shape == (shape[1] * shape[2] * shape[3], 2)
    x = x.reshape(shape[2], shape[3], shape[1], 2).permute(2, 0, 1, 3)
    vnnlib[0] = (x.reshape(-1, 2).numpy().tolist(), vnnlib[0][1])
    print('Model converted to NCHW format:', model_ori)
    return model_ori, vnnlib


def conv_output_shape(h_w, kernel_size=1, stride=1, pad=0, dilation=1):
    """
    Utility function for computing output of convolutions
    takes a tuple of (h,w) and returns a tuple of (h,w)
    """
    if type(h_w) is not tuple:
        h_w = (h_w, h_w)
    if type(kernel_size) is not tuple:
        kernel_size = (kernel_size, kernel_size)
    if type(stride) is not tuple:
        stride = (stride, stride)
    if type(pad) is not tuple:
        pad = (pad, pad)
    h = (h_w[0] + (2 * pad[0]) - (dilation * (kernel_size[0] - 1)) - 1) // stride[0] + 1
    w = (h_w[1] + (2 * pad[1]) - (dilation * (kernel_size[1] - 1)) - 1) // stride[1] + 1
    return h, w


def transpose_linear_layers(model_ori):
    """
    For test models in VNNComp.
    It looks like `in_features` and `out_features` are in the wrong order  after converting the onnx model to pytorch model.
    Return swapped model.
    """

    modules = []
    for m in model_ori._modules.values():
        if isinstance(m, nn.Linear):
            layer = nn.Linear(m.in_features, m.out_features)  # Fix a bug in onnx converter for test models.
            layer.weight.data = m.weight.data.to(torch.float)
            layer.bias.data = m.bias.data.to(torch.float) if m.bias is not None else torch.zeros_like(layer.bias.data)
            modules.append(layer)
        else:
            modules.append(m)
    model_ori = nn.Sequential(*modules)
    print('Linear layers are transposed in this model.')

    return model_ori


def convert_carvana_model_vnnlib(model, vnnlib, c_mode='naive'):
    """
    We provide three ways to convert vnnlib in Carvana.
    naive: return the raw vnnlib
    one_by_one: return the vnnlib which try to verify all properties one by one (similar with standard verified accuracy)
    together: return the vnnlib which try to verify all properties at the same time (universal perturbation)
    """
    gt = np.array(vnnlib[0][0][-31*47:])  # binary mask ground truth
    new_x = vnnlib[0][0][:3*31*47]  # real input image
    assert np.all((gt[:, 0] == gt[:, 1]))   # make sure the mask ground truth has no perturbation
    assert len(np.unique(gt)) == 2  # make sure the mask ground truth is binary

    new_c = []
    if c_mode == 'one_by_one':
        # generate a spec which try to verify all properties one by one (similar with standard cifar-10 verified accuracy)
        for idx, gt_i in enumerate(gt):
            this_c = np.zeros((1, 2914))  # 2914 = 2*31*47, output size
            if gt_i[0] == 0:  # dim 0 > dim 1
                this_c[0, idx] = 1  # ground truth idx
                this_c[0, idx + 1457] = -1  # target idx
            else:  # dim 0 < dim 1
                this_c[0, idx] = -1  # target idx
                this_c[0, idx + 1457] = 1  # ground truth idx
            new_c.append((this_c, np.array([0])))
        model = nn.Sequential(model, nn.Flatten())

    elif c_mode == 'together':
        # generate a spec which try to verify all properties together
        this_c = np.zeros((1457, 2914))  # 2914 = 2*31*47, output size
        for idx, gt_i in enumerate(gt):
            if gt_i[0] == 0:  # dim 0 > dim 1
                this_c[idx, idx] = 1  # ground truth idx
                this_c[idx, idx + 1457] = -1  # target idx
            else:  # dim 0 < dim 1
                this_c[idx, idx] = -1  # target idx
                this_c[idx, idx + 1457] = 1  # ground truth idx
        new_c.append((this_c, np.zeros([1457])))
        model = nn.Sequential(model, nn.Flatten())

    elif c_mode == 'naive':
        # original spec, count correct classified pixels
        new_c.append(vnnlib[0][1][0])
        model = Step_carvana(model, gt[:, 0])
    else:
        raise NotImplementedError

    new_vnnlib = [(new_x, new_c)]
    return model, new_vnnlib


def split_carvana(path):
    # upsample: Conv_7; Conv_82
    # simp: Conv_7; Conv_13

    model_name = path

    onnx_model = onnx.load(model_name)

    if ("simp" in model_name):
        st_node, ed_node = "Conv_7", "Conv_13"
    elif ("upsample" in model_name):
        st_node, ed_node = "Conv_7", "Conv_82"
    else:
        raise NotImplementedError

    new_initializers = []
    new_nodes = []

    initializers = {}
    for onnx_module in onnx_model.graph.initializer:
        initializers[onnx_module.name] = nh.to_array(onnx_module)

    trigger = False
    for i, node in enumerate(onnx_model.graph.node):
        if (node.name == st_node): trigger = True
        if (trigger == True):
            for init_name in node.input:
                if (init_name not in initializers): continue
                conv_initializer = create_initializer_tensor(
                    name=init_name,
                    tensor_array=initializers[init_name],
                    data_type=onnx.TensorProto.FLOAT
                )
                new_initializers.append(conv_initializer)
            new_nodes.append(node)
        if (node.name == ed_node): break

    model_input_name = new_nodes[0].input[0]
    X = onnx.helper.make_tensor_value_info(model_input_name,
                                           onnx.TensorProto.FLOAT,
                                           [None, 3, 31, 47])
    model_output_name = new_nodes[-1].output[0]
    Y = onnx.helper.make_tensor_value_info(model_output_name,
                                           onnx.TensorProto.FLOAT,
                                           [None, 2, 31, 47])

    new_graph = onnx.helper.make_graph(
        nodes=new_nodes,
        name="SplitNet",
        inputs=[X],
        outputs=[Y],
        initializer=onnx_model.graph.initializer
    )

    model_def = onnx.helper.make_model(new_graph, producer_name="onnx_example")

    model_def.opset_import[0].version = 12
    model_def = onnx.shape_inference.infer_shapes(model_def)

    onnx.checker.check_model(model_def)
    new_model_name = model_name[:-5] + "_split.onnx"
    onnx.save(model_def, new_model_name)
    print('New Carvana model saved to', new_model_name)


def customized_NN4SYS_loader(file_root, onnx_path, vnnlib_path):
    """
    Customized NN4SYS loader.
    We split the model into v1 and v2 models to resolve numerical issues
    """

    def get_path(onnx_file):
        return f'{os.path.basename(onnx_file)}.pt'

    def convert_and_save_nn4sys(onnx_file):
        model_ori, _ = load_model_onnx(onnx_file)
        model_ori = nn.Sequential(*list(model_ori._modules.values()))
        # Split the model into v1 and v2 models to resolve numerical issues
        modules_v1 = []
        modules_v2 = []
        stage = 1
        for m in model_ori._modules.values():
            if isinstance(m, nn.Linear):
                if m.weight.abs().max() > 1e9:
                    stage = 2 if len(modules_v2) == 0 else 3
                    continue
            else:
                continue
            if stage == 1:
                modules_v1 += [m, nn.ReLU(inplace=True)]
            elif stage == 2:
                dim = modules_v1[-2].out_features - 1
                lin = nn.Linear(m.in_features - dim, m.out_features - dim)
                lin.weight.data = m.weight[:lin.out_features, :lin.in_features]
                lin.weight.data = lin.weight.to(dtype=torch.float64)
                lin.bias.data = m.bias[:lin.out_features]
                lin.bias.data = lin.bias.to(dtype=torch.float64)
                modules_v2 += [lin, nn.ReLU(inplace=True)]
        x = torch.tensor([[119740.8]], dtype=torch.float64)
        modules_v1 = modules_v1[:-1]
        model_v1 = nn.Sequential(*modules_v1)
        y = model_v1(x)
        dim = y.size(-1) - 1
        modules_v2 = modules_v2[:-1]
        linear_ident = nn.Linear(1, dim, bias=False)
        linear_ident.weight.data = torch.ones_like(linear_ident.weight, dtype=torch.float64)
        modules_v2.insert(0, linear_ident)
        model_v2 = nn.Sequential(*modules_v2)
        y[:, :-2] *= (y[:, 1:-1] <= 0).int()
        select = (y[:, :-1] > 0).int()
        y2 = model_v2(x)
        y2 = y2[:] * select
        res = y2.sum(dim=-1, keepdim=True)
        res_ref = model_ori(x)
        model_ori = (model_v1, model_v2, model_ori)
        name = get_path(onnx_file)
        torch.save(model_ori, name)
        print(f'Converted model saved to {name}')

    shape = (-1, 1)
    path = get_path(os.path.join(file_root, onnx_path))
    if not os.path.exists(path):
        convert_and_save_nn4sys(os.path.join(file_root, onnx_path))
    # load pre-converted model
    model_ori = torch.load(path)
    print(f'Loaded from {path}')
    vnnlib = read_vnnlib(os.path.join(file_root, vnnlib_path), regression=True)
    return model_ori, shape, vnnlib


def customized_Carvana_loader(file_root, onnx_path, vnnlib_path):
    """
    Customized Carvana loader.
    We split the model for verification necessarily part only.
    """
    shape = (-1, 3, 31, 47)
    path = os.path.join(file_root, onnx_path[:-5] + '_split.onnx')
    if not os.path.exists(path):
        print('Split carvana model from:', os.path.join(file_root, onnx_path))
        split_carvana(os.path.join(file_root, onnx_path))
    else:
        print(f'Loaded split model from {path}')
    model_ori, _ = load_model_onnx(path)
    vnnlib = read_vnnlib(os.path.join(file_root, vnnlib_path))
    model_ori, vnnlib = convert_carvana_model_vnnlib(model_ori, vnnlib, c_mode='naive')
    return model_ori, shape, vnnlib


def customized_Marabou_loader(file_root, onnx_path, vnnlib_path):
    """
    Customized Marabou loader.
    Transpose Conv layers from NHWC to NCHW format.
    Also transpose vnnlib in the same way.
    """
    shape = (-1, 3, 32, 32)
    model_ori, _ = load_model_onnx(os.path.join(file_root, onnx_path))
    conv_c, conv_h, conv_w = shape[1:]
    modules = list(model_ori.modules())[1:]
    new_modules = []
    for mi, m in enumerate(modules):
        if isinstance(m, torch.nn.Conv2d):
            # Infer the output size of conv.
            conv_h, conv_w = conv_output_shape((conv_h, conv_w), m.kernel_size, m.stride, m.padding)
            # conv_c = m.weight.size(0)
        if isinstance(m, onnx2pytorch.operations.reshape.Reshape):
            # Replace reshape with flatten.
            new_modules.append(nn.Flatten())
        elif isinstance(m, torch.nn.ReLU) and mi == (len(modules)-1):
            # not add relu if last layer is relu
            pass
        else:
            new_modules.append(m)
    model_ori = nn.Sequential(*new_modules)
    vnnlib = read_vnnlib(os.path.join(file_root, vnnlib_path))
    # Transpose vnnlib when NHWC is used (marabou-cifar10).
    model_ori, vnnlib = transpose_nhwc(model_ori, vnnlib, shape)
    return model_ori, shape, vnnlib


def customized_Gtrsb_loader(file_root, onnx_path, vnnlib_path):
    """
    Customized GTRSB loader.
    Transpose vnnlib from NHWC to NCHW format.
    """
    model_ori, onnx_shape = load_model_onnx(os.path.join(file_root, onnx_path))
    shape = (1, *onnx_shape)
    vnnlib = read_vnnlib(os.path.join(file_root, vnnlib_path))
    x = torch.tensor(vnnlib[0][0])
    x = x.reshape(shape[2], shape[3], shape[1], 2).permute(2, 0, 1, 3)
    vnnlib[0] = (x.reshape(-1, 2).numpy().tolist(), vnnlib[0][1])
    model_ori, vnnlib = joint_optimization_with_onnx_and_vnnlib(model_ori, shape, vnnlib,
                                                                arguments.Config['model']['onnx_vnnlib_joint_optimization_flags'])
    return model_ori, shape, vnnlib


def customized_TEST_loader(file_root, onnx_path, vnnlib_path):
    """
    Customized test model loader in VNNComp.
    The linear layers in this model seems in the wrong order.
    """
    model_ori, onnx_shape = load_model_onnx(os.path.join(file_root, onnx_path))
    shape = (-1, *onnx_shape)  # add the batch dim to onnx_shape
    vnnlib = read_vnnlib(os.path.join(file_root, vnnlib_path))
    model_ori = transpose_linear_layers(model_ori)
    return model_ori, shape, vnnlib



#########################################################################
##   This file is part of the α,β-CROWN (alpha-beta-CROWN) verifier    ##
##                                                                     ##
##   Copyright (C) 2021-2024 The α,β-CROWN Team                        ##
##   Primary contacts: Huan Zhang <huan@huan-zhang.com>                ##
##                     Zhouxing Shi <zshi@cs.ucla.edu>                 ##
##                     Kaidi Xu <kx46@drexel.edu>                      ##
##                                                                     ##
##    See CONTRIBUTORS for all author contacts and affiliations.       ##
##                                                                     ##
##     This program is licensed under the BSD 3-Clause License,        ##
##        contained in the LICENCE file in this directory.             ##
##                                                                     ##
#########################################################################
r"""
This file is an example of defining a custom operation and providing its
relaxations for bound computation. Here we consider a modified ReLU
function which is a mixture between a linear function and ReLU function:
             / a_i x_i + b_i   if m_i = 1,
    f(x_i) = |
             \ ReLU(x_i)       if m_i = 0.
where a, b are element-wise slopes and biases when the function is linear,
and m is the mask controlling the behavior of this function. We consider
perturbations on x.

An example command to run verification on this customized model:

python abcrown.py --config exp_configs/tutorial_examples/custom_op_example.yaml --complete_verifier skip --batch_size 256

Note that if you also want to conduct branch and bound on your customized
op, you may also need to customize BaB code, so the complete verifier is
skipped here.
"""
import torch
import torch.nn as nn
from auto_LiRPA import register_custom_op
from auto_LiRPA.bound_ops import BoundRelu, Interval


class LinearMaskedReluOp(torch.autograd.Function):
    """A relu function with some neurons replaced with linear operations."""
    @staticmethod
    def forward(ctx, input: torch.Tensor, mask: torch.Tensor, slope: torch.Tensor, bias: torch.Tensor) -> torch.Tensor:
        # mask = 1 => using linear operation input * slope + bias, mask = 0 => using ReLU
        ctx.save_for_backward(input, mask, slope, bias)
        return input.clamp(min=0) * (1.0 - mask) + (input * slope + bias) * mask

    def backward(ctx, grad_output):
        """
        In the backward pass we receive a Tensor containing the gradient of the loss
        with respect to the output, and we need to compute the gradient of the loss
        with respect to the input.
        """
        input, mask, slope, bias = ctx.saved_tensors
        relu_grad = grad_output.clone()
        relu_grad[input < 0] = 0
        grad_input = relu_grad * (1.0 - mask) + grad_output * mask * slope
        grad_slope = grad_output * input * mask
        grad_bias = grad_output * mask
        grad_mask = -input.clamp(min=0) + (input * slope + bias)
        return grad_input, grad_mask, grad_slope, grad_bias

    @staticmethod
    def symbolic(g, input, mask, weight, bias):
        # This will be parsed as a custom operation when doing the ONNX conversion.
        return g.op("customOp::LinearMaskedRelu", input, mask, weight, bias)


class LinearMaskedRelu(nn.Module):
    """Create a module to wrap the parameters for LinearMaskedReluOp."""
    def __init__(self, size):
        super().__init__()
        if isinstance(size, int):
            size = (size, )
        # All mask, slope and bias are element-wise.
        self.register_buffer('mask', (torch.rand(size=size) > 0).to(dtype=torch.get_default_dtype()))
        self.register_buffer('alpha', torch.rand(size=size, dtype=torch.get_default_dtype()))
        self.register_buffer('bias', torch.rand(size=size, dtype=torch.get_default_dtype()))

    def forward(self, input):
        # mask = 1 => using linear operation input * slope + bias, mask = 0 => using ReLU
        return LinearMaskedReluOp.apply(input, self.mask, self.slope, self.bias)


class BoundLinearMaskedRelu(BoundRelu):
    """This class defines how we compute the bounds for our customized Relu function."""

    def forward(self, x, mask, slope, bias):
        """Regular forward propagation (e.g., for evaluating clean accuracy)."""
        # Save the shape, which will be used in other parts of the verifier.
        self.shape = x.shape[1:]
        if self.flattened_nodes is None:
            self.flattened_nodes = x[0].reshape(-1).shape[0]
        return LinearMaskedReluOp.apply(x, mask, slope, bias)

    def interval_propagate(self, x, mask, slope, bias):
        """Interval bound propagation (IBP)."""
        # Each x, mask, slope, bias is a tuple, or a Interval object representing lower and upper bounds.
        # We assume Linf norm perturbation on input.
        assert Interval.get_perturbation(x)[0] == float("inf")
        x_L, x_U = x[0], x[1]  # The inputs (x)
        # We assume no perturbations on mask, slope and bias.
        mask, slope, bias = mask[0], slope[0], bias[0]
        # Lower and upper bounds when ReLU is selected.
        relu_lb = x_L.clamp(min=0)
        relu_ub = x_U.clamp(min=0)
        # Lower and upper bounds when linear coefficients are selected.
        pos_slope = (slope >= 0).to(dtype=torch.get_default_dtype())
        neg_slope = 1.0 - pos_slope
        linear_lb = pos_slope * slope * x_L + neg_slope * slope * x_U + bias
        linear_ub = pos_slope * slope * x_U + neg_slope * slope * x_L + bias
        # Select the final bounds according to the mask.
        final_lb = mask * linear_lb + (1.0 - mask) * relu_lb
        final_ub = mask * linear_ub + (1.0 - mask) * relu_ub
        return final_lb, final_ub

    def _backward_relaxation(self, last_lA, last_uA, x, start_node, unstable_idx):
        """Element-wise CROWN relaxation for our special ReLU activation function."""
        # Call parent class to relax ReLU neurons.
        upper_d, upper_b, lower_d, lower_b, lb_lower_d, ub_lower_d, alpha_lookup_idx = super()._backward_relaxation(
                last_lA, last_uA, x, start_node, unstable_idx)
        # Modify the relaxation coefficients for these linear neurons.
        neg_mask = 1.0 - self._mask
        masked_slope = self._mask * self._slope
        masked_bias = self._mask * self._bias
        upper_d = masked_slope + neg_mask * upper_d
        upper_b = masked_bias + neg_mask * upper_b
        if lower_d is not None:
            # Shared slope between lower and upper bounds.
            lower_d = masked_slope + neg_mask * lower_d
        else:
            # Not shared slopes: we have two set of slopes one for lA, one for uA.
            # One of them might be not necessary (None), if only lower or upper bound is computed.
            lb_lower_d = masked_slope + neg_mask * lb_lower_d if lb_lower_d is not None else None
            ub_lower_d = masked_slope + neg_mask * ub_lower_d if ub_lower_d is not None else None
        assert lower_b is None  # For ReLU, there is no lower bias (=0)
        # The required dimension is (batch, spec, C, H, W). The size of masked_bias is (C,H,W),
        # and we need to expand other dimensions.
        lower_b = masked_bias.unsqueeze(0).unsqueeze(0).expand(upper_b.size())
        return upper_d, upper_b, lower_d, lower_b, lb_lower_d, ub_lower_d, alpha_lookup_idx

    def bound_backward(self, last_lA, last_uA, x, mask, slope, bias, **kwargs):
        """Backward LiRPA (CROWN) bound propagation."""
        # These are additional variables that will be used in _backward_relaxation(), so we save them here.
        self._mask = mask.buffer  # These are registered as buffers; see class BoundBuffer.
        self._slope = slope.buffer
        self._bias = bias.buffer
        # The parent class will call _backward_relaxation() and obtain the relaxations,
        # and that's all we need; after obtaining linear relaxations for each neuron, other
        # parts of class BoundRelu can be reused.
        As, lbias, ubias = super().bound_backward(last_lA, last_uA, x, **kwargs)
        # Returned As = [(lA, uA)]; these A matrices are for input x.
        # Our customized ReLU has three additional buffers as inputs; we need to set their
        # corresponding A matrices to None. The length of As must match the number of inputs
        # of this customize function.
        As += [(None, None), (None, None), (None, None)]
        return As, lbias, ubias


class linear_masked_relu_model(nn.Module):
    """Model for testing."""
    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv2d(3, 16, 4, stride=2, padding=1)
        # Using our customized ReLU function to replace the original ReLU function.
        self.linear_masked_relu1 = LinearMaskedRelu(size=(16,16,16))
        self.conv2 = nn.Conv2d(16, 32, 4, stride=2, padding=1)
        self.linear_masked_relu2 = LinearMaskedRelu(size=(32,8,8))
        self.flatten = nn.Flatten()
        self.linear1 = nn.Linear(32*8*8,100)
        self.linear_masked_relu3 = LinearMaskedRelu(100)
        self.linear2 = nn.Linear(100, 10)
        # Register the customized op in auto_LiRPA library.
        register_custom_op("customOp::LinearMaskedRelu", BoundLinearMaskedRelu)

    def forward(self, x):
        out = self.conv1(x)
        out = self.linear_masked_relu1(out)
        # out = torch.nn.functional.relu(out)
        out = self.conv2(out)
        out = self.linear_masked_relu2(out)
        # out = torch.nn.functional.relu(out)
        out = self.flatten(out)  # Flatten must be after activation for most efficient computation.
        out = self.linear1(out)
        out = self.linear_masked_relu3(out)
        # out = torch.nn.functional.relu(out)
        out = self.linear2(out)
        return out




#########################################################################
##   This file is part of the α,β-CROWN (alpha-beta-CROWN) verifier    ##
##                                                                     ##
##   Copyright (C) 2021-2024 The α,β-CROWN Team                        ##
##   Primary contacts: Huan Zhang <huan@huan-zhang.com>                ##
##                     Zhouxing Shi <zshi@cs.ucla.edu>                 ##
##                     Kaidi Xu <kx46@drexel.edu>                      ##
##                                                                     ##
##    See CONTRIBUTORS for all author contacts and affiliations.       ##
##                                                                     ##
##     This program is licensed under the BSD 3-Clause License,        ##
##        contained in the LICENCE file in this directory.             ##
##                                                                     ##
#########################################################################
import argparse
import sys
import os

parser = argparse.ArgumentParser()

parser.add_argument("CATEGORY", type=str)
parser.add_argument("ONNX_FILE", type=str, default=None, help='ONNX_FILE')
parser.add_argument("VNNLIB_FILE", type=str, default=None, help='VNNLIB_FILE')
parser.add_argument("RESULTS_FILE", type=str, default=None, help='RESULTS_FILE')
parser.add_argument("TIMEOUT", type=float, default=180, help='timeout for one property')
parser.add_argument("--DEBUG", action='store_true', help='whether to run in debug mode (checking saved adv example)')
parser.add_argument("--NOPGD", action='store_true', help='do not use pdg attack')
parser.add_argument("--TRY_CROWN", action='store_true', help='overwrite bound-prop-method to CROWN to save memory')

args = parser.parse_args()

python_path = sys.executable
library_path = os.path.dirname(os.path.realpath(__file__))

cmd = f"{python_path} {library_path}/abcrown.py --config {library_path}/"


# vnncomp 2021
if args.CATEGORY == "cifar10_resnet":
    cmd += "/exp_configs/vnncomp21/cifar10-resnet.yaml"

elif args.CATEGORY == "eran":
    if 'SIGMOID' in args.ONNX_FILE:
        cmd += "exp_configs/vnncomp21/eran_sigmoid.yaml"
    else:
        cmd += "exp_configs/vnncomp21/eran_mlp.yaml"

elif args.CATEGORY == "marabou-cifar10":
    cmd += "exp_configs/vnncomp21/marabou_cifar10.yaml"

elif args.CATEGORY == "verivital":
    cmd += "exp_configs/vnncomp21/verivital.yaml"

# common
# elif args.CATEGORY == "acasxu":
#     cmd += "exp_configs/vnncomp22/acasxu.yaml"

elif args.CATEGORY == "cifar2020":
    cmd += "exp_configs/vnncomp22/cifar2020_2_255.yaml"

elif args.CATEGORY == "oval21":
    cmd += "exp_configs/vnncomp22/oval22.yaml"

elif args.CATEGORY in ["mnist_fc", "mnistfc"]:  # same benchmark with different names
    if '256x2' in args.ONNX_FILE:
        cmd += "exp_configs/vnncomp22/mnistfc_small.yaml"
    else:
        cmd += "exp_configs/vnncomp22/mnistfc.yaml"

# special case: both vnncomp 2021 and 2022 have nn4sys, but they are different
elif args.CATEGORY == "nn4sys":
    if 'lindex' in args.ONNX_FILE or '128d' in args.ONNX_FILE or '2048d' in args.ONNX_FILE or 'pensieve' in args.ONNX_FILE:
        # All models in 2022 and 2023 use the same config file.
        cmd += "exp_configs/vnncomp23/nn4sys_2023.yaml"
    else:
        # nn4sys in 2021
        cmd += "exp_configs/vnncomp21/nn4sys.yaml"

# vnncomp 2022
elif args.CATEGORY == "carvana_unet_2022":
    cmd += "exp_configs/vnncomp22/carvana-unet-all.yaml"

elif args.CATEGORY == "cifar100_tinyimagenet_resnet":
    if 'CIFAR100_resnet_small' in args.ONNX_FILE:
        cmd += "exp_configs/vnncomp22/cifar100_small_2022.yaml"
    elif 'CIFAR100_resnet_medium' in args.ONNX_FILE:
        cmd += "exp_configs/vnncomp22/cifar100_med_2022.yaml"
    elif 'CIFAR100_resnet_large' in args.ONNX_FILE:
        cmd += "exp_configs/vnncomp22/cifar100_large_2022.yaml"
    elif 'CIFAR100_resnet_super' in args.ONNX_FILE:
        cmd += "exp_configs/vnncomp22/cifar100_super_2022.yaml"
    else:
        cmd += "exp_configs/vnncomp22/tinyimagenet_2022.yaml"

elif args.CATEGORY == "cifar_biasfield":
    cmd += "exp_configs/vnncomp22/cifar_biasfield.yaml"

# elif args.CATEGORY == "collins_rul_cnn":
#     cmd += "exp_configs/vnncomp22/collins-rul-cnn.yaml"

elif args.CATEGORY == "oval21":
    cmd += "exp_configs/vnncomp22/oval22.yaml"

elif args.CATEGORY == "reach_prob_density":
    if 'gcas' in args.ONNX_FILE:
        cmd += "exp_configs/vnncomp22/reach_probability_gcas.yaml"
    else:
        cmd += "exp_configs/vnncomp22/reach_probability.yaml"

elif args.CATEGORY == "rl_benchmarks":
    if 'cartpole' in args.ONNX_FILE:
        cmd += "exp_configs/vnncomp22/cartpole.yaml"
    elif 'lunarlander' in args.ONNX_FILE:
        cmd += "exp_configs/vnncomp22/lunarlander.yaml"
    else:
        cmd += "exp_configs/vnncomp22/dubins-rejoin.yaml"

elif args.CATEGORY == "sri_resnet_a":
    cmd += "exp_configs/vnncomp22/resnet_A.yaml"

elif args.CATEGORY == "sri_resnet_b":
    cmd += "exp_configs/vnncomp22/resnet_B.yaml"

elif args.CATEGORY == "tllverifybench":
    cmd += "exp_configs/vnncomp22/tllVerifyBench.yaml"

elif args.CATEGORY == "vggnet16_2022":
    cmd += "exp_configs/vnncomp22/vggnet16.yaml"

# vnncomp 2023
elif args.CATEGORY == "acasxu":
    # caution: acasxu is actually a common benchmark across multiple years,
    # we have exactly the same config file in vnncomp23/ and vnncomp22/ folders
    # but for consistency, we point to the config in vnncomp23/ folder
    cmd += "exp_configs/vnncomp23/acasxu.yaml"

elif args.CATEGORY == "cctsdb_yolo":
    cmd += "exp_configs/vnncomp23/cctsdb_yolo.yaml"

elif args.CATEGORY == "cgan":
    cmd += "exp_configs/vnncomp23/cgan.yaml"

elif args.CATEGORY == "collins_rul_cnn":
    cmd += "exp_configs/vnncomp23/collins-rul-cnn.yaml"

elif args.CATEGORY == "collins_yolo_robustness":
    cmd += "exp_configs/vnncomp23/collins_yolo_robustness.yaml"

elif args.CATEGORY == "dist_shift":
    cmd += "exp_configs/vnncomp23/dist-shift.yaml"

elif args.CATEGORY == "metaroom":
    cmd += "exp_configs/vnncomp23/metaroom.yaml"

elif args.CATEGORY == "ml4acopf":
    cmd += "exp_configs/vnncomp23/ml4acopf.yaml"

elif args.CATEGORY == "tllverifybench":
    cmd += "exp_configs/vnncomp23/tllVerifyBench.yaml"

elif args.CATEGORY == "traffic_signs_recognition":
    cmd += "exp_configs/vnncomp23/gtrsb.yaml"

elif args.CATEGORY == "vggnet16":
    # same config as last year
    cmd += "exp_configs/vnncomp23/vggnet16.yaml"

elif args.CATEGORY == "vit":
    cmd += "exp_configs/vnncomp23/vit.yaml"

elif args.CATEGORY == "yolo":
    cmd += "exp_configs/vnncomp23/yolo-xiangruzh.yaml"

elif args.CATEGORY == "test":
    pass

else:
    exit("CATEGORY {} not supported yet".format(args.CATEGORY))

# test case may run in other args.CATEGORY at the end of them, so we parse them here to allow correct measurement of overhead.
if os.path.split(args.VNNLIB_FILE)[-1] in ['test_' + f + '.vnnlib' for f in ['nano', 'tiny', 'small']]:
    cmd = f"{python_path} {library_path}/abcrown.py --config {library_path}/exp_configs/vnncomp21/test.yaml"
elif 'test_prop' in args.VNNLIB_FILE:
    cmd = f"{python_path} {library_path}/abcrown.py --config {library_path}/exp_configs/vnncomp22/acasxu.yaml"


cmd += " --precompile_jit"
cmd += " --onnx_path " + str(args.ONNX_FILE)
cmd += " --vnnlib_path " + str(args.VNNLIB_FILE)
cmd += " --results_file " + str(args.RESULTS_FILE)
cmd += " --timeout " + str(args.TIMEOUT)

# save adv example to args.RESULTS_FILE
cmd += " --save_adv_example"

# use CROWN bound propagation, when original run triggers OOM, run_instance.sh will add this flag
if args.TRY_CROWN:
    # This also disables the use of output constraints. They are only useful for alpha-CROWN
    cmd += " --bound_prop_method crown --apply_output_constraints_to"

# verify the adv example everytime it's saved
if args.DEBUG:
    cmd += " --eval_adv_example"

# do not use pdg attack during verification
if args.NOPGD:
    cmd += " --pgd_order=skip"

print("\n------------------------- COMMAND ------------------------------")
print(cmd)
print("----------------------------------------------------------------\n")

ret = os.system(cmd)
if ret != 0:
    # avoid original return code to be > 255, reserve its non-zero feature
    sys.exit(int(ret) % 255 + 1)



#########################################################################
##   This file is part of the α,β-CROWN (alpha-beta-CROWN) verifier    ##
##                                                                     ##
##   Copyright (C) 2021-2024 The α,β-CROWN Team                        ##
##   Primary contacts: Huan Zhang <huan@huan-zhang.com>                ##
##                     Zhouxing Shi <zshi@cs.ucla.edu>                 ##
##                     Kaidi Xu <kx46@drexel.edu>                      ##
##                                                                     ##
##    See CONTRIBUTORS for all author contacts and affiliations.       ##
##                                                                     ##
##     This program is licensed under the BSD 3-Clause License,        ##
##        contained in the LICENCE file in this directory.             ##
##                                                                     ##
#########################################################################
import numpy as np
import torch.nn.functional as F

def customized_gtrsb_saver(adv_example, adv_output, res_path):
    # almost the same as the original save_cex function
    # permute the input back before flattening the tensor
    # (See customized_Gtrsb_loader() from custom/custom_model_loader.py

    adv_example = adv_example.permute(0, 1, 3, 4, 2).contiguous()

    x = adv_example.view(-1).detach().cpu()
    adv_output = F.softmax(adv_output).detach().cpu().numpy()
    with open(res_path, 'w+') as f:
        input_dim = np.prod(adv_example[0].shape)
        f.write("(")
        for i in range(input_dim):
            f.write("(X_{}  {})\n".format(i, x[i].item()))

        for i in range(adv_output.shape[1]):
            if i == 0:
                f.write("(Y_{} {})".format(i, adv_output[0,i]))
            else:
                f.write("\n(Y_{} {})".format(i, adv_output[0,i]))
        f.write(")")
        f.flush()

